"""CoBib logging module."""

import logging
import logging.config


def log_to_stream(level='WARNING'):
    """Configures a StreamHandler logger.

    Args:
        level (str, optional): verbosity level indicator.
    """
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(name)s %(funcName)s:%(lineno)d %(message)s'
            },
        },
        'handlers': {
            'default': {
                'formatter': 'standard',
                'class': 'logging.StreamHandler',
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': level,
                'formatter': 'standard',
                'propagate': True
            }
        }
    })


def log_to_file(level='INFO', logfile='/tmp/cobib.log'):
    """Configures a FileHandler logger.

    Args:
        level (str, optional): verbosity level indicator.
        logfile (str, optional): output path for log file.
    """
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s [%(levelname)s] %(name)s %(funcName)s:%(lineno)d %(message)s'
            },
        },
        'handlers': {
            'default': {
                'formatter': 'standard',
                'class': 'logging.FileHandler',
                'filename': logfile,
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': level,
                'formatter': 'standard',
                'propagate': True
            }
        }
    })



"""CoBib's TUI state."""

import logging

from enum import Enum

from cobib.config import CONFIG

LOGGER = logging.getLogger(__name__)


class Mode(Enum):
    """The possible view modes."""
    LIST = 'list'
    SHOW = 'show'
    SEARCH = 'search'


class State:
    """State class to track the stateful parameters of CoBib's TUI.

    State objects are used to store all stateful parameters of CoBib's TUI and simplify the handling
    of these parameters across the TUI and Frame objects.
    """

    def __init__(self):
        """Initializes the State object."""
        LOGGER.debug('Initializing the State')
        self.top_line = 0
        self.left_edge = 0
        self.current_line = 0
        self.previous_line = -1

        self.mode = Mode.LIST.value
        self.inactive_commands = []
        self.topstatus = ''

        # these cannot be set yet, because the CONFIG has not been fully populated at the time of
        # creation of the STATE singleton
        self.list_args = []

    def initialize(self):
        """Initialize configuration-dependent settings."""
        self.list_args = CONFIG.config['TUI'].get('default_list_args').split(' ')
        if CONFIG.config['TUI'].getboolean('reverse_order', True):
            self.list_args += ['-r']

    def update(self, state):
        """Update from other state.

        Args:
            state (State): another (deep-copied) state.
        """
        self.top_line = state.top_line
        self.left_edge = state.left_edge
        self.current_line = state.current_line
        self.previous_line = state.previous_line

        self.mode = state.mode
        self.inactive_commands = state.inactive_commands
        self.topstatus = state.topstatus

        self.list_args = state.list_args


STATE = State()



"""CoBib redo command."""

import argparse
import logging
import os
import subprocess
import sys

from cobib.config import CONFIG
from cobib.database import read_database
from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class RedoCommand(Command):
    """Redo Command."""

    name = 'redo'

    def execute(self, args, out=sys.stdout):
        """Redo last undone change.

        Redoes the last undone change to the database file.

        Args: See base class.
        """
        conf_database = CONFIG.config['DATABASE']
        git_tracked = conf_database.getboolean('git')
        if not git_tracked:
            msg = "You must enable CoBib's git-tracking in order to use the `Redo` command. " + \
                "Please refer to the man-page for more information on how to do so."
            print(msg, file=sys.stderr)
            LOGGER.error(msg)
            return

        file = os.path.realpath(os.path.expanduser(conf_database['file']))
        root = os.path.dirname(file)
        if not os.path.exists(os.path.join(root, '.git')):
            msg = "You have configured, but not initialized CoBib's git-tracking. " + \
                "Please consult `cobib init --help` for more information on how to do so."
            print(msg, file=sys.stderr)
            LOGGER.error(msg)
            return

        LOGGER.debug('Starting Redo command.')
        parser = ArgumentParser(prog="redo", description="Redo subcommand parser.")

        try:
            # pylint: disable=unused-variable
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        LOGGER.debug('Obtaining git log.')
        lines = subprocess.check_output([
            "git", "--no-pager", "-C", f"{root}", "log", "--oneline", "--no-decorate", "--no-abbrev"
        ])
        redone_shas = set()
        for commit in lines.decode().strip().split('\n'):
            LOGGER.debug('Processing commit %s', commit)
            sha, *message = commit.split()
            if message[0] == 'Redo':
                # Store already redone commit sha
                LOGGER.debug('Storing undone commit sha: %s', sha)
                redone_shas.add(message[-1])
                continue
            if sha in redone_shas:
                LOGGER.info('Skipping %s as it was already redone', sha)
                continue
            LOGGER.debug('Attempting to redo %s.', sha)
            commands = [
                f"git -C {root} revert --no-commit {sha}",
                f"git -C {root} commit --no-gpg-sign --quiet --message 'Redo {sha}'"
            ]
            redo = subprocess.Popen('; '.join(commands), shell=True,
                                    stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            redo.communicate()
            if redo.returncode != 0:
                LOGGER.error('Redo was unsuccessful. Please consult the logs and git history of'
                             ' your database for more information.')
            break
        else:
            msg = "Could not find a commit to redo. Please commit something first!"
            print(msg, file=sys.stderr)
            LOGGER.warning(msg)
            sys.exit(1)

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Redo command triggered from TUI.')
        tui.execute_command(['redo'], skip_prompt=True)
        # update database list
        LOGGER.debug('Updating list after Redo command.')
        read_database(fresh=True)
        tui.viewport.update_list()
        # if cursor line is below buffer height, move it one line back up
        if tui.STATE.current_line >= tui.viewport.buffer.height:
            tui.STATE.current_line -= 1



"""CoBib show command."""

import argparse
import logging
import sys

from cobib import __version__
from cobib.config import CONFIG
from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class ShowCommand(Command):
    """Show Command."""

    name = 'show'

    def execute(self, args, out=sys.stdout):
        """Show entry.

        Prints the details of a selected entry in BibLaTex format to stdout.

        Args: See base class.
        """
        LOGGER.debug('Starting Show command.')
        parser = ArgumentParser(prog="show", description="Show subcommand parser.")
        parser.add_argument("label", type=str, help="label of the entry")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        try:
            entry = CONFIG.config['BIB_DATA'][largs.label]
            entry_str = entry.to_bibtex()
            print(entry_str, file=out)
        except KeyError:
            msg = f"No entry with the label '{largs.label}' could be found."
            LOGGER.error(msg)
            print(msg, file=out)

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Show command triggered from TUI.')
        # get current label
        label, cur_y = tui.viewport.get_current_label()
        # populate buffer with entry data
        LOGGER.debug('Clearing current buffer contents.')
        tui.viewport.clear()
        ShowCommand().execute([label], out=tui.viewport.buffer)
        tui.viewport.buffer.split()
        if label in tui.selection:
            LOGGER.debug('Current entry is selected. Applying highlighting.')
            tui.viewport.buffer.replace(0, label,
                                        CONFIG.get_ansi_color('selection') + label + '\x1b[0m')
        LOGGER.debug('Populating buffer with ShowCommand result.')
        tui.viewport.view(ansi_map=tui.ANSI_MAP)

        # reset current cursor position
        tui.STATE.top_line = 0
        tui.STATE.current_line = 0
        # update top statusbar
        tui.STATE.topstatus = "CoBib v{} - {}".format(__version__, label)
        tui.statusbar(tui.topbar, tui.STATE.topstatus)
        # enter show menu
        tui.STATE.mode = 'show'
        tui.STATE.previous = cur_y
        tui.STATE.inactive_commands = ['Add', 'Filter', 'Show', 'Sort']



"""CoBib's TUI viewport."""

import copy
import curses
import logging
import re

from cobib import __version__
from cobib.commands import ListCommand
from cobib.config import CONFIG
from .buffer import TextBuffer
from .state import Mode, STATE

LOGGER = logging.getLogger(__name__)


class Frame:
    """Frame class used to integrate a `TextBuffer` and `curses.pad` tightly with each other.

    This helper class is mainly used to implement the main 'viewport' of the TUI. It is a stateful
    object which handles a session-persistent history of the viewport to enable seamless level
    navigation.
    """

    def __init__(self, tui, max_height, max_width):
        """Initializes the Frame object."""
        LOGGER.debug("Initializing frame's buffer")
        self.buffer = TextBuffer()
        LOGGER.debug("Initializing frame's pad")
        self.pad = curses.newpad(1, 1)
        # Also store a history of buffer contents and state (such as the current line, etc.)
        self.history = []

        # store TUI reference
        self.tui = tui
        self.height = max_height
        self.width = max_width

    def clear(self):
        """Wrapper for buffer.clear to intercept for history storage."""
        self.history.append((copy.deepcopy(self.buffer), copy.deepcopy(STATE)))
        # almost 100 entries should be much more than enough
        self.history = self.history[:99]
        self.buffer.clear()

    def revert(self):
        """Reverts the frame to the previous state."""
        if not self.history:
            LOGGER.debug('Empty frame history, nothing to revert')
            return
        self.buffer, state = self.history.pop()
        STATE.update(state)
        self.buffer.replace(range(self.buffer.height), re.escape(CONFIG.get_ansi_color('selection'))
                            + r'(.+)' + re.escape('\x1b[0m'), r'\1')
        # highlight current selection
        for label in self.tui.selection:
            # Note: this step may become a performance bottleneck because we replace inside the
            # whole buffer for each selected label!
            if STATE.mode == Mode.SEARCH.value:
                # Note: the inclusion of the search label is explained in the `SearchCommand`.
                self.buffer.replace(range(self.buffer.height),
                                    re.escape(CONFIG.get_ansi_color('search_label')) + label
                                    + re.escape('\x1b[0m'),
                                    CONFIG.get_ansi_color('search_label') +
                                    CONFIG.get_ansi_color('selection') + label + '\x1b[0m\x1b[0m')
            else:
                # Note: the two spaces are explained in the `select()` method.
                self.buffer.replace(range(self.buffer.height), label + '  ',
                                    CONFIG.get_ansi_color('selection') + label + '\x1b[0m  ')
        self.view(ansi_map=self.tui.ANSI_MAP)
        self.tui.statusbar(self.tui.topbar, STATE.topstatus)

    def resize(self, new_height, new_width):
        """Resizes the Frame to the new dimensions.

        Args:
            new_height (int): the new height.
            new_width (int): the new width.
        """
        self.height = new_height
        self.width = new_width
        self.refresh()

    def refresh(self):
        """Utility function to quickly refresh the Frame's pad."""
        self.pad.refresh(STATE.top_line, STATE.left_edge, 1, 0, self.height, self.width-1)

    def view(self, ansi_map=None):
        """Utility function to quickly view the Frame's buffer.

        Args:
            ansi_map (dict): optional, dictionary mapping ANSI codes to curses color pairs.
        """
        self.buffer.view(self.pad, self.height, self.width-1, ansi_map=ansi_map)

    def scroll_y(self, update):
        """Scroll vertically.

        Args:
            update (int or str): the offset specifying the scrolling height.
        """
        scrolloff = CONFIG.config['TUI'].getint('scroll_offset', 3)
        overlap = scrolloff >= self.height - scrolloff
        scroll_lock = overlap and STATE.current_line - STATE.top_line == self.height // 2
        # jump to top
        if update == 'g':
            LOGGER.debug('Jump to top of viewport.')
            STATE.top_line = 0
            STATE.current_line = 0
        # jump to bottom
        elif update == 'G':
            LOGGER.debug('Jump to bottom of viewport.')
            STATE.top_line = max(self.buffer.height - self.height, 0)
            STATE.current_line = self.buffer.height-1
        # scroll up
        elif update < 0:
            LOGGER.debug('Scroll viewport up by %d lines.', update)
            next_line = STATE.current_line + update
            if STATE.top_line > 0 and next_line < STATE.top_line + scrolloff:
                if scroll_lock or not overlap:
                    STATE.top_line += update
                elif overlap and \
                        STATE.current_line - STATE.top_line > self.height // 2 and \
                        next_line - STATE.top_line < self.height // 2:
                    STATE.top_line = next_line - self.height // 2
            STATE.current_line = max(next_line, 0)
        # scroll down
        elif update > 0:
            LOGGER.debug('Scroll viewport down by %d lines.', update)
            next_line = STATE.current_line + update
            if next_line >= STATE.top_line + self.height - scrolloff and \
                    self.buffer.height > STATE.top_line + self.height:
                if scroll_lock or not overlap:
                    STATE.top_line += update
                elif overlap and \
                        STATE.current_line - STATE.top_line < self.height // 2 and \
                        next_line - STATE.top_line > self.height // 2:
                    STATE.top_line = next_line - self.height // 2
            if next_line < self.buffer.height:
                STATE.current_line = next_line
            else:
                STATE.top_line = self.buffer.height - self.height
                STATE.current_line = self.buffer.height - 1

    def scroll_x(self, update):
        """Scroll horizontally.

        Args:
            update (int or str): the offset specifying the scrolling width.
        """
        # jump to beginning
        if update == 0:
            LOGGER.debug('Jump to first column of viewport.')
            STATE.left_edge = 0
        # jump to end
        elif update == '$':
            LOGGER.debug('Jump to end of viewport.')
            STATE.left_edge = self.buffer.width - self.width
        else:
            LOGGER.debug('Scroll viewport horizontally by %d columns.', update)
            next_col = STATE.left_edge + update
            # limit column such that no empty columns can appear on left or right
            if 0 <= next_col <= self.buffer.width - self.width:
                STATE.left_edge = next_col

    def wrap(self):
        """Toggles wrapping of the text currently displayed."""
        LOGGER.debug('Wrap command triggered.')
        # first, ensure left_edge is set to 0
        STATE.left_edge = 0
        # then, wrap the buffer
        self.buffer.wrap(self.width)
        self.view()
        # if cursor line is below buffer height, move it one line back up
        if self.buffer.height and STATE.current_line >= self.buffer.height:
            STATE.current_line -= 1

    def get_current_label(self):
        """Returns the label and y position of the currently selected entry."""
        LOGGER.debug('Obtaining current label "under" cursor.')
        cur_y, _ = self.pad.getyx()
        # Two cases are possible: the list and the show mode
        if STATE.mode == Mode.LIST.value:
            # In the list mode, the label can be found in the current line
            # or in one of the previous lines if we are on a wrapped line
            while chr(self.pad.inch(cur_y, 0)) == TextBuffer.INDENT[0]:
                cur_y -= 1
            label = self.pad.instr(cur_y, 0).decode('utf-8').split(' ')[0]
        elif STATE.mode == Mode.SEARCH.value:
            # In the search mode, the same holds but we need to slightly change the label detection.
            while chr(self.pad.inch(cur_y, 0)) in ('[', TextBuffer.INDENT[0]):
                cur_y -= 1
            label = self.pad.instr(cur_y, 0).decode('utf-8').split(' ')[0]
        else:
            # In any other mode, the label can be found in the top statusbar
            label = '-'.join(STATE.topstatus.split('-')[1:]).strip()
            # We also set cur_y to 0 for the select command to find it
            cur_y = 0
        LOGGER.debug('Current label at "%s" is "%s".', str(cur_y), label)
        return label, cur_y

    def update_list(self):
        """Updates the default list view."""
        LOGGER.debug('Re-populating the viewport with the list command.')
        self.buffer.clear()
        labels = ListCommand().execute(STATE.list_args, out=self.buffer)
        labels = labels or []  # convert to empty list if labels is None
        # populate buffer with the list
        if STATE.mode != Mode.LIST.value:
            STATE.current_line = STATE.previous_line
            STATE.mode = Mode.LIST.value
        # reset viewport
        STATE.top_line = 0
        STATE.left_edge = 0
        STATE.inactive_commands = []
        # highlight current selection
        for label in self.tui.selection:
            # Note: the two spaces are explained in the `select()` method.
            # Also: this step may become a performance bottleneck because we replace inside the
            # whole buffer for each selected label!
            self.buffer.replace(range(self.buffer.height), label + '  ',
                                CONFIG.get_ansi_color('selection') + label + '\x1b[0m  ')
        # display buffer in viewport
        self.view(ansi_map=self.tui.ANSI_MAP)
        # update top statusbar
        STATE.topstatus = "CoBib v{} - {} Entries".format(__version__, len(labels))
        self.tui.statusbar(self.tui.topbar, STATE.topstatus)
        # if cursor position is out-of-view (due to e.g. top-line reset in Show command), reset the
        # top-line such that the current line becomes height again
        if STATE.current_line > STATE.top_line + self.height:
            STATE.top_line = min(STATE.current_line, self.buffer.height - self.height)



"""Cobib init command."""

import argparse
import logging
import os
import sys

from cobib.config import CONFIG
from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class InitCommand(Command):
    """Init Command."""

    name = 'init'

    def execute(self, args, out=sys.stdout):
        """Initialize database.

        Initializes the yaml database file at the configured location.

        Args: See base class.
        """
        LOGGER.debug('Starting Init command.')
        parser = ArgumentParser(prog="init", description="Init subcommand parser.")
        parser.add_argument('-g', '--git', action='store_true',
                            help="initialize git repository")

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        conf_database = CONFIG.config['DATABASE']
        file = os.path.realpath(os.path.expanduser(conf_database['file']))
        root = os.path.dirname(file)

        file_exists = os.path.exists(file)
        git_tracked = os.path.exists(os.path.join(root, '.git'))

        if file_exists:
            if git_tracked:
                msg = 'Database file already exists and is being tracked by git. ' + \
                      'There is nothing else to do.'
                print(msg, file=sys.stderr)
                LOGGER.info(msg)
                return

            if not git_tracked and not largs.git:
                msg = 'Database file already exists! Use --git to start tracking it with git.'
                print(msg, file=sys.stderr)
                LOGGER.warning(msg)
                return

        else:
            LOGGER.debug('Creating path for database file: "%s"', root)
            os.makedirs(root, exist_ok=True)

            LOGGER.debug('Creating empty database file: "%s"', file)
            open(file, 'w').close()

        if largs.git:
            if not conf_database.getboolean('git'):
                msg = 'You are about to initialize the git tracking of your database, but this ' + \
                      'will only have effect if you also enable the DATABASE/git setting in ' + \
                      'your configuration file!'
                print(msg, file=sys.stderr)
                LOGGER.warning(msg)
            # First, check whether git is configured correctly.
            print('Checking `git config --get user.name`:', end=' ', flush=True)
            name_set = os.waitstatus_to_exitcode(os.system('git config --get user.name'))
            print()
            print('Checking `git config --get user.email`:', end=' ', flush=True)
            email_set = os.waitstatus_to_exitcode(os.system('git config --get user.email'))
            print()
            if name_set != 0 or email_set != 0:
                msg = 'In order to use git you must configure your name and email first! For ' + \
                      'more information please consult `man gittutorial`.'
                print(msg, file=sys.stderr)
                LOGGER.warning(msg)
                sys.exit(1)
            LOGGER.debug('Initializing git repository in "%s"', root)
            os.system(f'git init {root}')
            self.git(args=vars(largs), force=True)



"""CoBib Command interface."""

import argparse
import json
import logging
import os
import shlex
import sys
from abc import ABC, abstractmethod

from cobib.config import CONFIG

LOGGER = logging.getLogger(__name__)


class Command(ABC):
    """The Command interface declares a method for command execution and some helper methods."""

    name = 'base'

    @abstractmethod
    def execute(self, args, out=sys.stdout):
        """Command execution.

        Args:
            args (dict): additional arguments used for the execution.
            out (stream, optional): possible alternative to stdout.
        """

    @staticmethod
    def tui(tui):
        """TUI command interface.

        Args:
            tui (cobib.TUI): instance of CoBib's TUI.
        """

    def git(self, args=None, force=False):
        """Track command's changes with git.

        Args:
            args (optional, dict): a dictionary containing the command arguments.
            force (boolean): whether to ignore the configuration setting.
        """
        conf_database = CONFIG.config['DATABASE']
        git_tracked = conf_database.getboolean('git')
        if not git_tracked and not force:
            return

        file = os.path.realpath(os.path.expanduser(conf_database['file']))
        root = os.path.dirname(file)

        if not os.path.exists(os.path.join(root, '.git')):
            if git_tracked:
                msg = 'You have configured CoBib to track your database with git. ' + \
                      'Please run `cobibt init --git`, to initialize this tracking.'
                print(msg, file=sys.stderr)
                LOGGER.warning(msg)
                return

        msg = f"Auto-commit: {self.name.title()}Command"
        if args:
            msg += '\n\n'
            msg += json.dumps(args, indent=2, default=str)

        commands = [
            f'cd {root}',
            f'git add -- {file}',
            f'git commit --no-gpg-sign --quiet --message {shlex.quote(msg)}',
        ]
        LOGGER.debug('Auto-commit to git from %s command.', self.name)
        os.system('; '.join(commands))


class ArgumentParser(argparse.ArgumentParser):
    """Overwrite ArgumentParser to allow catching any error messages thrown by parse_args.

    Source: https://stackoverflow.com/a/5943381
    """

    def _get_action_from_name(self, name):
        """Given a name, get the Action instance registered with this parser.

        If only it were made available in the ArgumentError object. It is
        passed as it's first argument...

        Args:
            name (str or None): name of the action.
        """
        container = self._actions
        if name is None:
            return None
        for action in container:
            if '/'.join(action.option_strings) == name:
                return action
            if action.metavar == name:
                return action
            if action.dest == name:
                return action
        return None

    def error(self, message):
        """Prints an error.

        Args:
            message (str): error message string.
        """
        exc = sys.exc_info()[1]
        if exc:
            raise exc
        super().error(message)



"""CoBib curses interface."""

import curses
import logging
import re
import shlex
import sys
from functools import partial
from signal import signal, SIGWINCH

from cobib import commands
from cobib.config import CONFIG
from .buffer import TextBuffer, InputBuffer
from .frame import Frame
from .state import Mode, STATE

LOGGER = logging.getLogger(__name__)


class TUI:
    """CoBib's curses-based TUI.

    The TUI is implemented as a class to simplify management of different windows/pads and keep a
    synchronized state most consistently.
    """

    COLOR_VALUES = {
        'black': curses.COLOR_BLACK,
        'blue': curses.COLOR_BLUE,
        'cyan': curses.COLOR_CYAN,
        'green': curses.COLOR_GREEN,
        'magenta': curses.COLOR_MAGENTA,
        'red': curses.COLOR_RED,
        'white': curses.COLOR_WHITE,
        'yellow': curses.COLOR_YELLOW,
    }

    COLOR_NAMES = [
        'top_statusbar',
        'bottom_statusbar',
        'search_label',
        'search_query',
        'cursor_line',
        'popup_help',
        'popup_stdout',
        'popup_stderr',
        'selection',
    ]

    ANSI_MAP = {}

    # available command dictionary
    COMMANDS = {
        'Add': commands.AddCommand.tui,
        'Delete': commands.DeleteCommand.tui,
        'Edit': commands.EditCommand.tui,
        'Export': commands.ExportCommand.tui,
        'Filter': partial(commands.ListCommand.tui, sort_mode=False),
        'Help': lambda self: self.help(),
        'Modify': commands.ModifyCommand.tui,
        'Open': commands.OpenCommand.tui,
        'Prompt': lambda self: self.execute_command(None),
        'Quit': lambda self: self.quit(),
        'Redo': commands.RedoCommand.tui,
        'Search': commands.SearchCommand.tui,
        'Select': lambda self: self.select(),
        'Show': commands.ShowCommand.tui,
        'Sort': partial(commands.ListCommand.tui, sort_mode=True),
        'Undo': commands.UndoCommand.tui,
        'Wrap': lambda self: self.viewport.wrap(),
        'x': lambda self, update: self.viewport.scroll_x(update),
        'y': lambda self, update: self.viewport.scroll_y(update),
    }

    # command help strings
    HELP_DICT = {
        "Add": "Prompts for a new entry to be added to the database.",
        "Delete": "Removes the current entry from the database.",
        "Edit": "Edits the current entry in an external EDITOR.",
        "Export": "Allows exporting the database to .bib or .zip files.",
        "Filter": "Allows filtering the list via `++/--` keywords.",
        "Help": "Displays this help.",
        "Modify": "Allows basic modification of multiple entries at once.",
        "Open": "Opens the associated file of an entry.",
        "Prompt": "Executes arbitrary CoBib CLI commands in the prompt.",
        "Quit": "Closes current menu and quit's CoBib.",
        "Redo": "Redoes the last undone change. Requires git-tracking!",
        "Search": "Searches the database for a given string.",
        "Select": "Adds the current entry to the interactive selection.",
        "Show": "Shows the details of an entry.",
        "Sort": "Prompts for the field to sort against (-r to reverse).",
        "Undo": "Undoes the last change. Requires git-tracking!",
        "Wrap": "Wraps the text displayed in the window.",
    }

    # standard key bindings
    KEYDICT = {
        10: 'Show',  # line feed = ENTER
        13: 'Show',  # carriage return = ENTER
        curses.KEY_DOWN: ('y', 1),
        curses.KEY_UP: ('y', -1),
        curses.KEY_NPAGE: ('y', 20),
        curses.KEY_PPAGE: ('y', -20),
        ord('j'): ('y', 1),
        ord('k'): ('y', -1),
        ord('g'): ('y', 'g'),
        ord('G'): ('y', 'G'),
        2: ('y', -20),  # CTRL-B
        4: ('y', 10),  # CTRL-D
        6: ('y', 20),  # CTRL-F
        21: ('y', -10),  # CTRL-U
        curses.KEY_LEFT: ('x', -1),
        curses.KEY_RIGHT: ('x', 1),
        ord('h'): ('x', -1),
        ord('l'): ('x', 1),
        ord('0'): ('x', 0),
        ord('$'): ('x', '$'),
        ord(':'): 'Prompt',
        ord('/'): 'Search',
        ord('?'): 'Help',
        ord('a'): 'Add',
        ord('d'): 'Delete',
        ord('e'): 'Edit',
        ord('f'): 'Filter',
        ord('m'): 'Modify',
        ord('o'): 'Open',
        ord('q'): 'Quit',
        ord('r'): 'Redo',
        ord('s'): 'Sort',
        ord('u'): 'Undo',
        ord('v'): 'Select',
        ord('w'): 'Wrap',
        ord('x'): 'Export',
    }

    def __init__(self, stdscr):
        """Initializes the curses-TUI and starts the event loop.

        Args:
            stdscr (curses.window): the curses standard screen as returned by curses.initscr().
        """
        LOGGER.info('Initializing TUI.')
        self.stdscr = stdscr

        # register resize handler
        signal(SIGWINCH, self.resize_handler)

        # Clear and refresh the screen for a blank canvas
        self.stdscr.clear()
        self.stdscr.refresh()
        # Initialize layout
        curses.curs_set(0)
        self.height, self.width = self.stdscr.getmaxyx()
        LOGGER.debug('stdscr size determined to be %dx%d', self.width, self.height)
        # and colors
        LOGGER.debug('Initializing colors.')
        curses.use_default_colors()
        TUI.colors()
        # and user key mappings
        LOGGER.debug('Initializing key bindings.')
        TUI.bind_keys()

        # Initialize STATE
        LOGGER.debug('Initializing global State')
        self.STATE = STATE  # pylint: disable=invalid-name
        STATE.initialize()
        # load further configuration settings
        self.prompt_before_quit = CONFIG.config['TUI'].getboolean('prompt_before_quit', True)

        # the selection needs to be tracked outside of the State in order to persist across
        # different views
        self.selection = set()

        # Initialize top status bar
        LOGGER.debug('Populating top status bar.')
        self.topbar = curses.newwin(1, self.width, 0, 0)
        self.topbar.bkgd(' ', curses.color_pair(TUI.COLOR_NAMES.index('top_statusbar') + 1))

        # Initialize bottom status bar
        # NOTE: -2 leaves an additional empty line for the command prompt
        LOGGER.debug('Populating bottom status bar.')
        self.botbar = curses.newwin(1, self.width, self.height-2, 0)
        self.botbar.bkgd(' ', curses.color_pair(TUI.COLOR_NAMES.index('bottom_statusbar') + 1))
        self.statusbar(self.botbar, self.infoline())

        # Initialize prompt line
        # The prompt is a pad to allow command/error prompts to exceed the terminal width.
        self.prompt = curses.newpad(1, self.width)

        # Initialize main viewport
        LOGGER.debug('Initializing viewport with Frame')
        # NOTE: -3 accounts for the top and bottom statusline as well as the command prompt
        self.viewport = Frame(self, self.height-3, self.width)
        # populate buffer with list of reference entries
        LOGGER.debug('Populating viewport buffer.')
        self.viewport.update_list()

        # start key event loop
        LOGGER.debug('Starting key event loop.')
        self.loop()
        LOGGER.info('Exiting TUI.')

    def resize_handler(self, signum, frame):  # pylint: disable=unused-argument
        """Handles terminal window resize events.

        Args:
            signum (int): signal number.
            frame: unused argument, required by the function template.
        """
        LOGGER.debug('Handling resize event.')
        # stop curses window
        curses.endwin()
        # clear and refresh for a blank canvas
        self.stdscr.clear()
        self.stdscr.refresh()
        # update total dimension data
        self.height, self.width = self.stdscr.getmaxyx()
        LOGGER.debug('New stdscr dimension determined to be %dx%d', self.width, self.height)
        # update top statusbar
        self.topbar.resize(1, self.width)
        self.statusbar(self.topbar, STATE.topstatus)
        self.topbar.refresh()
        # update bottom statusbar
        self.botbar.resize(1, self.width)
        self.botbar.mvwin(self.height-2, 0)
        self.statusbar(self.botbar, self.infoline())
        self.botbar.refresh()
        # update prompt
        self.prompt.resize(1, self.width)
        self.prompt.refresh(0, 0, self.height-1, 0, self.height, self.width-1)
        # update viewport
        self.viewport.resize(self.height-3, self.width)

    def quit(self):
        """Breaks the key event loop or quits one viewport level."""
        if STATE.mode == Mode.LIST.value:
            LOGGER.debug('Quitting from lowest level.')
            if self.prompt_before_quit:
                msg = 'Do you really want to quit CoBib? [y/n] '
                curses.curs_set(1)
                self.prompt.clear()
                if len(msg) >= self.prompt.getmaxyx()[1] - 2:
                    self.prompt.resize(1, len(msg) + 2)
                self.prompt.insstr(0, 0, msg)
                self.prompt.move(0, len(msg))
                self.prompt.refresh(0, 0, self.height-1, 0, self.height, self.width-1)
                key = 0
                while True:
                    if key in (ord('y'), ord('Y')):
                        raise StopIteration
                    if key in (ord('n'), ord('N')):
                        LOGGER.info('User aborted quitting.')
                        curses.curs_set(0)
                        break
                    key = self.prompt.getch()
                self.prompt.clear()
                self.prompt.refresh(0, 0, self.height-1, 0, self.height, self.width-1)
            else:
                raise StopIteration
        LOGGER.debug('Quitting higher menu level. Falling back to list view.')
        self.viewport.revert()

    @staticmethod
    def colors():
        """Initialize the color pairs for the curses TUI."""
        # Start colors in curses
        curses.start_color()
        # parse user color configuration
        color_cfg = CONFIG.config['COLORS']
        colors = {col: {} for col in TUI.COLOR_NAMES}
        for attr, col in color_cfg.items():
            if attr in TUI.COLOR_VALUES.keys():
                if not curses.can_change_color():
                    # cannot change curses default colors
                    LOGGER.warning('Curses cannot change the default colors. Skipping color setup.')
                    continue
                # update curses-internal color with HEX-color
                rgb_color = tuple(int(col.strip('#')[i:i+2], 16) for i in (0, 2, 4))
                # curses colors range from 0 to 1000
                curses_color = tuple(col * 1000 // 255 for col in rgb_color)
                curses.init_color(TUI.COLOR_VALUES[attr], *curses_color)
            else:
                if attr[:-3] not in TUI.COLOR_NAMES:
                    LOGGER.warning('Detected unknown TUI color name specification: %s', attr[:-3])
                    continue
                colors[attr[:-3]][attr[-2:]] = col

        # initialize color pairs for TUI elements
        for idx, attr in enumerate(TUI.COLOR_NAMES):
            foreground = colors[attr].get('fg', 'white')
            background = colors[attr].get('bg', 'black')
            LOGGER.debug('Initiliazing color pair %d for %s', idx + 1, attr)
            curses.init_pair(idx + 1, TUI.COLOR_VALUES[foreground], TUI.COLOR_VALUES[background])
            LOGGER.debug('Adding ANSI color code for %s', attr)
            TUI.ANSI_MAP[CONFIG.get_ansi_color(attr)] = TUI.COLOR_NAMES.index(attr) + 1

    @staticmethod
    def bind_keys():
        """Bind keys according to user configuration."""
        key_bindings = CONFIG.config['KEY_BINDINGS']
        for command, key in key_bindings.items():
            LOGGER.info('Binding key %s to the %s command.', key, command)
            if command not in TUI.COMMANDS.keys():
                LOGGER.warning('Unknown command "%d". Ignoring key binding.', command)
                continue
            if key == 'ENTER':
                TUI.KEYDICT[10] = command  # line feed
                TUI.KEYDICT[13] = command  # carriage return
                continue
            if isinstance(key, str):
                # map key to its ASCII number
                key = ord(key)
            TUI.KEYDICT[key] = command

    @staticmethod
    def statusbar(statusline, text, attr=0):
        """Update the text in the provided status bar and refresh it.

        Args:
            statusline (curses.window): single line height window used as a statusline.
            text (str): text to place in the statusline.
            attr (int, optional): attribute number to use for the printed text.
        """
        statusline.erase()
        _, max_x = statusline.getmaxyx()
        statusline.addnstr(0, 0, text, max_x-1, attr)
        statusline.refresh()

    @staticmethod
    def infoline():
        """Returns a list of the available key bindings."""
        infoline = ''
        for cmd in TUI.HELP_DICT:
            if cmd:
                # get associated key for this command
                for key, command in TUI.KEYDICT.items():
                    if cmd == command:
                        key = 'ENTER' if key in (10, 13) else chr(key)
                        infoline += " {}:{}".format(key, cmd)
                        break
            else:
                infoline += "  "
        return infoline.strip()

    def help(self):
        """Help command.

        Opens a new curses window with more detailed information on the configured key bindings and
        short descriptions of the commands.
        """
        LOGGER.debug('Help command triggered.')
        # populate text buffer with help text
        help_text = TextBuffer()
        LOGGER.debug('Generating help text.')
        for cmd, desc in TUI.HELP_DICT.items():
            key = ' '
            for key, command in TUI.KEYDICT.items():
                if cmd == command:
                    # find mapped key
                    key = 'ENTER' if key in (10, 13) else chr(key)
                    break
            # write: [key] Command: Description
            help_text.write("{:^8} {:<8} {}".format('['+key+']', cmd+':', desc))
        # add header section
        help_text.lines.insert(0, "{0:^{1}}".format("CoBib TUI Help", help_text.width))
        help_text.lines.insert(1, "{:^8} {:<8} {}".format('Key', 'Command', 'Description'))
        help_text.height += 2

        # open help popup
        help_text.popup(self, background=TUI.COLOR_NAMES.index('popup_help'))

    def loop(self):
        """The key-handling event loop."""
        key = 0
        # key is the last character pressed
        while True:
            # handle possible keys
            try:
                if key in TUI.KEYDICT.keys():
                    cmd = TUI.KEYDICT[key]
                    if cmd not in STATE.inactive_commands:
                        if isinstance(cmd, tuple):
                            TUI.COMMANDS[cmd[0]](self, cmd[1])
                        else:
                            TUI.COMMANDS[cmd](self)
                elif key == curses.KEY_RESIZE:
                    self.resize_handler(None, None)
            except StopIteration:
                LOGGER.debug('Stopping key event loop.')
                # raised by quit command
                break

            # highlight current line
            current_colors = []
            for x_pos in range(0, self.viewport.pad.getmaxyx()[1]):
                x_ch = self.viewport.pad.inch(STATE.current_line, x_pos)
                current_colors.append(x_ch)
                # x_ch is the character at the queried position. The bottom 8 bits are the character
                # proper, and upper bits are the attributes.
                # Source: https://docs.python.org/3/library/curses.html#curses.window.inch
                # Thus, we can find the color attribute by striping the last 8 bits.
                x_attr = x_ch >> 8
                if x_attr <= TUI.COLOR_NAMES.index('cursor_line'):
                    self.viewport.pad.chgat(STATE.current_line, x_pos, 1, curses.color_pair(
                        TUI.COLOR_NAMES.index('cursor_line') + 1))

            # Refresh the screen
            self.viewport.refresh()

            # Wait for next input
            key = self.stdscr.getch()
            LOGGER.debug('Key press registered: %s', str(key))

            # reset highlight of current line
            for x_pos in range(0, self.viewport.pad.getmaxyx()[1]):
                self.viewport.pad.chgat(STATE.current_line, x_pos, 1, current_colors[x_pos])

    def select(self):
        """Toggles selection of the current label."""
        LOGGER.debug('Select command triggered.')
        # get current label
        label, cur_y = self.viewport.get_current_label()
        # toggle selection
        if label not in self.selection:
            LOGGER.info("Adding '%s' to the selection.", label)
            self.selection.add(label)
            # Note, that we use an additional two spaces to attempt to uniquely identify the label
            # in the list mode. Otherwise it might be possible that the same text (as used for the
            # label) can occur elsewhere in the buffer.
            # We do not need this outside of the list view because then the line indexed by `cur_y`
            # will surely only include the one label which we actually want to operate on.
            offset = '  ' if STATE.mode == Mode.LIST.value else ''
            self.viewport.buffer.replace(cur_y, label + offset, CONFIG.get_ansi_color('selection')
                                         + label + '\x1b[0m' + offset)
        else:
            LOGGER.info("Removing '%s' from the selection.", label)
            self.selection.remove(label)
            self.viewport.buffer.replace(cur_y, re.escape(CONFIG.get_ansi_color('selection'))
                                         + label + re.escape('\x1b[0m'), label)
        # update buffer view
        self.viewport.view(ansi_map=self.ANSI_MAP)

    def prompt_print(self, text):
        """Handle printing text to the prompt line.

        This function also handles a bug in curses which disallows 'newline'-characters [1].
        Thus, if this is the case, the first line is printed in the prompt and the total text is
        also presented in a popup window. The reason for the duplicate information is too provide a
        little context on the previous message once the popup window has been closed.

        [1] https://docs.python.org/3/library/curses.html#curses.window.addstr

        Args:
            text (str or list): the test to print to the prompt.
        """
        lines = text.strip().split('\n') if isinstance(text, str) else text
        self.prompt.clear()
        self.prompt.resize(1, max(len(lines[0]), self.width))
        self.prompt.addstr(0, 0, lines[0])
        self.prompt.refresh(0, 0, self.height-1, 0, self.height, self.width-1)
        if len(lines) > 1:
            buffer = TextBuffer()
            buffer.write(text)
            buffer.popup(self, background=TUI.COLOR_NAMES.index('popup_stderr'))

    def prompt_handler(self, command, symbol=":"):
        """Handle prompt input.

        Args:
            command (str or None): the command string to populate the prompt with.
            symbol (str, optional): the prompt symbol.

        Returns:
            The final user command.
        """
        LOGGER.debug('Handling input by the user in prompt line.')
        # make cursor visible
        curses.curs_set(1)

        # populate prompt line and place cursor
        prompt_line = symbol if command is None else f"{symbol}{command} "
        self.prompt.clear()
        self.prompt.resize(1, max(len(prompt_line)+2, self.width))
        self.prompt.addstr(0, 0, prompt_line)
        self.prompt.move(0, len(prompt_line))
        self.prompt.refresh(0, 0, self.height-1, 0, self.height, self.width-1)

        key = 0
        command = ''
        while True:
            # get current cursor position
            _, cur_x = self.prompt.getyx()
            # get next key
            self.prompt.nodelay(False)
            key = self.prompt.getch()
            # handle keys
            if key == 27:  # ESC
                self.prompt.nodelay(True)
                # check if it was an arrow escape sequence
                self.prompt.getch()
                arrow = self.prompt.getch()
                if arrow == -1:
                    LOGGER.debug('Prompt input aborted by the user.')
                    # if not, ESC ends the prompt
                    break
                if arrow == 68:  # left arrow key
                    LOGGER.debug('Move cursor left from arrow key input.')
                    self.prompt.move(_, cur_x - 1)
                elif arrow == 67:  # right arrow key
                    LOGGER.debug('Move cursor right from arrow key input.')
                    self.prompt.move(_, cur_x + 1)
            elif key == 127:  # BACKSPACE
                if cur_x > 1:
                    LOGGER.debug('Delete the previous character in the prompt.')
                    self.prompt.delch(_, cur_x - 1)
                else:
                    LOGGER.debug('Prompt input aborted by the user.')
                    command = ''
                    break
            elif key in (10, 13):  # ENTER
                LOGGER.debug('Execute the command as prompted.')
                command = self.prompt.instr(0, 1).decode('utf-8').strip()
                break
            else:
                # any normal key is simply echoed
                self.prompt.resize(1, max(cur_x + 2, self.width))
                self.prompt.addstr(_, cur_x, chr(key))
                self.prompt.move(_, cur_x + 1)
            self.prompt.refresh(0, max(0, cur_x - self.width + 2),
                                self.height-1, 0, self.height, self.width-1)
        # leave echo mode and make cursor invisible
        curses.curs_set(0)

        # clear prompt line
        self.prompt.clear()
        self.prompt.refresh(0, 0, self.height-1, 0, self.height, self.width-1)

        return command

    def execute_command(self, command, out=None, pass_selection=False, skip_prompt=False):
        """Executes a command.

        Args:
            command (str or None): the command to execute.
            out (stream, optional): the output stream to redirect stdout to.
            pass_selection (boolean, optional): whether to the pass the current TUI selection in the
                                                executed command arguments.
            skip_prompt (boolean, optional): whether to skip the user prompt for further commands.

        Returns:
            A pair with the first element being the list with the executed command to allow further
            handling and the second element being whatever is returned by the command.
        """
        if not skip_prompt:
            command = self.prompt_handler(command)
            # split command into separate arguments for cobib
            command = shlex.split(command)

        # process command if it non empty and actually has arguments
        result = None
        if command and command[0]:
            LOGGER.debug('Processing the command: %s', ' '.join(command))
            # temporarily disable prints to stdout, stderr and stdin
            original_stdin = sys.stdin
            original_stdout = sys.stdout
            original_stderr = sys.stderr
            sys.stdout = TextBuffer()
            sys.stderr = TextBuffer()
            sys.stdin = InputBuffer(buffer=sys.stdout, tui=self)
            # run command
            subcmd = getattr(commands, command[0].title()+'Command')()
            try:
                if pass_selection:
                    command += ['--']
                    command.extend(list(self.selection))
                result = subcmd.execute(command[1:], out=out)
            except SystemExit:
                pass
            # if error occurred print info to prompt
            if sys.stderr.lines:
                LOGGER.warning('The command "%s" resulted in an error.', ' '.join(command))
                sys.stderr.split()
                LOGGER.info('sys.stderr contains:\n%s', '\n'.join(sys.stderr.lines))
                # wrap before checking the height:
                sys.stderr.wrap(self.width)
                if sys.stderr.height > 1:
                    sys.stderr.popup(self, background=TUI.COLOR_NAMES.index('popup_stderr'))
                else:
                    self.prompt_print(sys.stderr.lines)
                # command exited with an error
                self.viewport.update_list()
            elif sys.stdout.lines:
                LOGGER.info('A message to stdout from "%s" was intercepted.', ' '.join(command))
                sys.stdout.split()
                LOGGER.info('sys.stdout contains:\n%s', '\n'.join(sys.stdout.lines))
                # wrap before checking the height:
                sys.stdout.wrap(self.width)
                if sys.stdout.height > 1:
                    sys.stdout.popup(self, background=TUI.COLOR_NAMES.index('popup_stdout'))
                else:
                    self.prompt_print(sys.stdout.lines)
            # restore stdout, stderr and stdin
            sys.stdout = original_stdout
            sys.stderr = original_stderr
            sys.stdin = original_stdin
        # return command to enable additional handling by function caller
        return (command, result)



"""CoBib open command."""

import argparse
import logging
import os
import subprocess
import sys
from collections import defaultdict
from urllib.parse import urlparse

from cobib.config import CONFIG
from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class OpenCommand(Command):
    """Open Command."""

    name = 'open'

    # pylint: disable=too-many-branches
    def execute(self, args, out=sys.stderr):
        """Open file from entries.

        Opens the associated file of the entries with xdg-open.

        Args: See base class.
        """
        LOGGER.debug('Starting Open command.')
        parser = ArgumentParser(prog="open", description="Open subcommand parser.")
        parser.add_argument("labels", type=str, nargs='+', help="labels of the entries")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        # pylint: disable=too-many-nested-blocks
        for label in largs.labels:
            things_to_open = defaultdict(list)
            count = 0
            # first: find all possible things to open
            try:
                entry = CONFIG.config['BIB_DATA'][label]
                for field in ('file', 'url'):
                    if field in entry.data.keys() and entry.data[field]:
                        for val in entry.data[field].split(','):
                            val = val.strip()
                            LOGGER.debug('Parsing "%s" for URLs.', val)
                            things_to_open[field] += [urlparse(val)]
                            count += 1
            except KeyError:
                msg = "Error: No entry with the label '{}' could be found.".format(label)
                LOGGER.error(msg)
                continue

            # if there are none, skip current label
            if not things_to_open:
                msg = "Warning: This entry has no actionable field associated with it."
                LOGGER.warning(msg)
                print(msg, file=out or sys.stderr)
                continue

            if count == 1:
                # we found a single URL to open
                self._open_url(list(things_to_open.values())[0][0])
            else:
                # we query the user what to do
                idx = 1
                url_list = []
                prompt = []
                # print formatted list of available URLs
                for field, urls in things_to_open.items():
                    for url in urls:
                        prompt.append(f"{idx:3}: [{field}] {url.geturl()}")
                        url_list.append(url)
                        idx += 1
                # loop until the user picks a valid choice
                help_requested = False
                while True:
                    prompt_copy = prompt.copy()
                    prompt_copy.append("Entry to open [Type 'help' for more info]: ")
                    try:
                        choice = input('\n'.join(prompt_copy)).strip()
                    except EOFError:
                        choice = ''
                    if not choice:
                        # empty input
                        msg = 'User aborted open command.'
                        LOGGER.warning(msg)
                        print(msg, file=sys.stderr)
                        break
                    if choice == 'help':
                        LOGGER.debug('User requested help.')
                        if not help_requested:
                            msg = ["You can specify one of the following options:",
                                   "  1. a url number",
                                   "  2. a field name provided in '[...]'",
                                   "  3. or simply 'all'",
                                   "  4. ENTER will abort the command",
                                   ""]
                            prompt = msg + prompt
                        help_requested = True
                    elif choice == 'all':
                        LOGGER.debug('User selected all urls.')
                        for url in url_list:
                            self._open_url(url)
                        break
                    elif choice in things_to_open.keys():
                        LOGGER.debug('User selected the %s set of urls.', choice)
                        for url in things_to_open[choice]:
                            self._open_url(url)
                        break
                    elif choice.isdigit() and int(choice) > 0 and int(choice) <= count:
                        LOGGER.debug('User selected url %s', choice)
                        self._open_url(url_list[int(choice)-1])
                        break

    @staticmethod
    def _open_url(url):
        """Opens a URL."""
        opener = CONFIG.config['DATABASE'].get('open', None)
        try:
            url = url.geturl() if url.scheme else os.path.abspath(url.geturl())
            LOGGER.debug('Opening "%s" with %s.', url, opener)
            with open(os.devnull, 'w') as devnull:
                subprocess.Popen([opener, url], stdout=devnull, stderr=devnull,
                                 stdin=devnull, close_fds=True)
        except FileNotFoundError as err:
            LOGGER.error(err)
            print(err, file=sys.stderr)

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Open command triggered from TUI.')
        if tui.selection:
            # use selection for command
            labels = list(tui.selection)
        else:
            # get current label
            label, _ = tui.viewport.get_current_label()
            labels = [label]
        tui.execute_command(['open'] + labels, skip_prompt=True)



"""CoBib undo command."""

import argparse
import logging
import os
import subprocess
import sys

from cobib.config import CONFIG
from cobib.database import read_database
from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class UndoCommand(Command):
    """Undo Command."""

    name = 'undo'

    def execute(self, args, out=sys.stdout):
        """Undo last change.

        Undoes the last change to the database file. By default, only auto-committed changes by
        CoBib will be undone. Use `--force` to undo other changes, too.

        Args: See base class.
        """
        conf_database = CONFIG.config['DATABASE']
        git_tracked = conf_database.getboolean('git')
        if not git_tracked:
            msg = "You must enable CoBib's git-tracking in order to use the `Undo` command. " + \
                "Please refer to the man-page for more information on how to do so."
            print(msg, file=sys.stderr)
            LOGGER.error(msg)
            return

        file = os.path.realpath(os.path.expanduser(conf_database['file']))
        root = os.path.dirname(file)
        if not os.path.exists(os.path.join(root, '.git')):
            msg = "You have configured, but not initialized CoBib's git-tracking. " + \
                "Please consult `cobib init --help` for more information on how to do so."
            print(msg, file=sys.stderr)
            LOGGER.error(msg)
            return

        LOGGER.debug('Starting Undo command.')
        parser = ArgumentParser(prog="undo", description="Undo subcommand parser.")
        parser.add_argument("-f", "--force", action='store_true',
                            help="allow undoing non auto-committed changes")

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        LOGGER.debug('Obtaining git log.')
        lines = subprocess.check_output([
            "git", "--no-pager", "-C", f"{root}", "log", "--oneline", "--no-decorate", "--no-abbrev"
        ])
        undone_shas = set()
        for commit in lines.decode().strip().split('\n'):
            LOGGER.debug('Processing commit %s', commit)
            sha, *message = commit.split()
            if message[0] == 'Undo':
                # Store already undone commit sha
                LOGGER.debug('Storing undone commit sha: %s', sha)
                undone_shas.add(message[-1])
                continue
            if sha in undone_shas:
                LOGGER.info('Skipping %s as it was already undone', sha)
                continue
            if largs.force or (message[0] == 'Auto-commit:' and message[-1] != 'InitCommand'):
                # we undo a commit if and only if:
                #  - the `force` argument is specified OR
                #  - the commit is an `auto-committed` change which is NOT from `InitCommand`
                LOGGER.debug('Attempting to undo %s.', sha)
                commands = [
                    f"git -C {root} revert --no-commit {sha}",
                    f"git -C {root} commit --no-gpg-sign --quiet --message 'Undo {sha}'"
                ]
                undo = subprocess.Popen('; '.join(commands), shell=True,
                                        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                undo.communicate()
                if undo.returncode != 0:
                    LOGGER.error('Undo was unsuccessful. Please consult the logs and git history of'
                                 ' your database for more information.')
                break
        else:
            msg = "Could not find a commit to undo. Please commit something first!"
            print(msg, file=sys.stderr)
            LOGGER.warning(msg)
            sys.exit(1)

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Undo command triggered from TUI.')
        tui.execute_command(['undo'], skip_prompt=True)
        # update database list
        LOGGER.debug('Updating list after Undo command.')
        read_database(fresh=True)
        tui.viewport.update_list()
        # if cursor line is below buffer height, move it one line back up
        if tui.STATE.current_line >= tui.viewport.buffer.height:
            tui.STATE.current_line -= 1



"""Database handler module."""

from collections import OrderedDict
from pathlib import Path
import logging
import os
import sys

from cobib.config import CONFIG
from cobib.parser import Entry

LOGGER = logging.getLogger(__name__)


def read_database(fresh=False):
    """Reads the database file.

    The YAML database file pointed to by the configuration file is read in and parsed. The data is
    stored as an OrderedDict in the global configuration object.

    Args:
        fresh (bool, optional): Forcefully reloads the bibliographic data.
    """
    if fresh and 'BIB_DATA' in CONFIG.config.keys():
        LOGGER.info('Update the bibliographic data in memory.')
        # delete data currently in memory
        del CONFIG.config['BIB_DATA']
    conf_database = CONFIG.config['DATABASE']
    file = os.path.expanduser(conf_database['file'])
    try:
        LOGGER.info('Loading database file: %s', file)
        CONFIG.config['BIB_DATA'] = Entry.from_yaml(Path(file))
    except AttributeError:
        LOGGER.debug('Initializing an empty database.')
        CONFIG.config['BIB_DATA'] = OrderedDict()
    except FileNotFoundError:
        LOGGER.critical("The database file %s does not exist! Please run `cobib init`!", file)
        sys.exit(1)


def write_database(entries):
    """Writes to the database file.

    Appends the bibliographic data of the provided entries to the database file. If a label already
    exists in the database, the corresponding entry is skipped.

    Args:
        entries (list[Entry]): list of new bibliography entries

    Returns:
        A list of the actually written entries.
    """
    if 'BIB_DATA' not in CONFIG.config.keys():
        # if no data in memory, read the database file (the case when using the CLI)
        LOGGER.info('Reading database file, before trying to write to it.')
        read_database()
    new_lines = []
    new_entries = []
    for label, entry in entries.items():
        if label in CONFIG.config['BIB_DATA'].keys():
            LOGGER.warning("Label %s already exists! Ignoring the new version.", label)
            continue
        string = entry.to_yaml()
        reduced = '\n'.join(string.splitlines())
        new_lines.append(reduced)
        new_entries.append(label)

    if new_lines:
        conf_database = CONFIG.config['DATABASE']
        file = os.path.expanduser(conf_database['file'])
        # append new lines to the database file
        with open(file, 'a') as bib:
            for line in new_lines:
                LOGGER.debug('Appending line to database file: %s', line)
                bib.write(line+'\n')

    return new_entries



"""CoBib modify command."""

import argparse
import logging
import os
import sys

from cobib.config import CONFIG
from .base_command import ArgumentParser, Command
from .list import ListCommand

LOGGER = logging.getLogger(__name__)


class ModifyCommand(Command):
    """Modify Command."""

    name = 'modify'

    @staticmethod
    def field_value_pair(string):
        """Utility method to assert the field-value pair argument type.

        Args:
            string (str): the argument string to check.
        """
        # try splitting the string into field and value, any errors will be handled by argparse
        field, value = string.split(':')
        return (field, value)

    def execute(self, args, out=sys.stdout):
        """Modify entries.

        Allows bulk modification of entries.

        Args: See base class.
        """
        LOGGER.debug('Starting Modify command.')
        parser = ArgumentParser(prog="modify", description="Modify subcommand parser.")
        parser.add_argument("modification", type=self.field_value_pair,
                            help="Modification to apply to the specified entries."
                            "\nThis argument must be a string formatted as <field>:<value> where "
                            "field can be any field of the entries and value can be any string "
                            "which should be placed in that field. Be sure to escape this "
                            "field-value pair properly, especially if the value contains spaces."
                            )
        parser.add_argument("-a", "--append", action="store_true",
                            help="Appends to the modified field rather than overwriting it.")
        parser.add_argument("-s", "--selection", action="store_true",
                            help="When specified, the `filter` argument will be interpreted as "
                            "a list of entry labels rather than arguments for the `list` command.")
        parser.add_argument('filter', nargs='+',
                            help="You can specify filters as used by the `list` command in order "
                            "to select a subset of labels to be modified. To ensure this works as "
                            "expected you should add the pseudo-argument '--' before the list of "
                            "filters. See also `list --help` for more information.")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_intermixed_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        out = open(os.devnull, 'w')
        if largs.selection:
            LOGGER.info('Selection given. Interpreting `filter` as a list of labels')
            labels = largs.filter
        else:
            LOGGER.debug('Gathering filtered list of entries to be modified.')
            labels = ListCommand().execute(largs.filter, out=out)

        field, value = largs.modification

        if largs.append:
            msg = 'The append-mode of the `modify` command has not been implemented yet.'
            print(msg)
            LOGGER.warning(msg)
            sys.exit(1)

        for label in labels:
            try:
                entry = CONFIG.config['BIB_DATA'][label]
                entry.data[field] = value

                conf_database = CONFIG.config['DATABASE']
                file = os.path.expanduser(conf_database['file'])
                with open(file, 'r') as bib:
                    lines = bib.readlines()
                entry_to_be_replaced = False
                with open(file, 'w') as bib:
                    for line in lines:
                        if line.startswith(label + ':'):
                            LOGGER.debug('Entry "%s" found. Starting to replace lines.', label)
                            entry_to_be_replaced = True
                            continue
                        if entry_to_be_replaced and line.startswith('...'):
                            LOGGER.debug('Reached end of entry "%s".', label)
                            entry_to_be_replaced = False
                            bib.writelines('\n'.join(entry.to_yaml().split('\n')[1:]))
                            continue
                        if not entry_to_be_replaced:
                            bib.write(line)

                msg = f"'{label}' was modified."
                print(msg)
                LOGGER.info(msg)
            except KeyError:
                print("Error: No entry with the label '{}' could be found.".format(label))

        self.git(args=vars(largs))

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Modify command triggered from TUI.')
        # handle input via prompt
        if tui.selection:
            tui.execute_command('modify -s', pass_selection=True)
        else:
            tui.execute_command('modify')



"""Tests for CoBib's parsing module."""

from os import path
from pathlib import Path
import pytest
from cobib import parser
from cobib.config import CONFIG

EXAMPLE_BIBTEX_FILE = 'test/example_entry.bib'
EXAMPLE_YAML_FILE = 'test/example_entry.yaml'

EXAMPLE_ENTRY_DICT = {
    'ENTRYTYPE': 'article',
    'ID': 'Cao_2019',
    'author': "Yudong Cao and Jonathan Romero and Jonathan P. Olson and Matthias Degroote and Peter"
              + " D. Johnson and M{\\'a}ria Kieferov{\\'a} and Ian D. Kivlichan and Tim Menke "
              + "and Borja Peropadre and Nicolas P. D. Sawaya and Sukin Sim and Libor Veis and "
              + "Al{\\'a}n Aspuru-Guzik",
    'doi': '10.1021/acs.chemrev.8b00803',
    'journal': 'Chemical Reviews',
    'month': '8',
    'number': '19',
    'pages': '10856--10915',
    'publisher': 'American Chemical Society ({ACS})',
    'title': 'Quantum Chemistry in the Age of Quantum Computing',
    'url': 'https://doi.org/10.1021%2Facs.chemrev.8b00803',
    'volume': '119',
    'year': '2019',
}


def test_entry_set_label():
    """Test label changing."""
    # this test may fail if the input dict is not copied
    CONFIG.set_config()
    entry = parser.Entry('article', EXAMPLE_ENTRY_DICT)
    entry.set_label = 'Cao2019'
    assert entry.label == 'Cao2019'
    assert entry.data['ID'] == 'Cao2019'


def test_entry_set_tags():
    """Test tags setting."""
    CONFIG.set_config()
    entry = parser.Entry('article', EXAMPLE_ENTRY_DICT)
    # NB: tags must be a list
    entry.set_tags = ['foo']
    assert entry.data['tags'] == 'foo'
    # '+' signs are stripped
    entry.set_tags = ['+foo']
    assert entry.data['tags'] == 'foo'
    # also multiple occurrences
    entry.set_tags = ['++foo']
    assert entry.data['tags'] == 'foo'
    # list works as expected
    entry.set_tags = ['foo', 'bar']
    assert entry.data['tags'] == 'foo, bar'
    entry.set_tags = ['+foo', 'bar']
    assert entry.data['tags'] == 'foo, bar'
    entry.set_tags = ['+foo', '+bar']
    assert entry.data['tags'] == 'foo, bar'


def test_entry_set_file():
    """Test file setting."""
    CONFIG.set_config()
    entry = parser.Entry('article', EXAMPLE_ENTRY_DICT)
    entry.set_file = EXAMPLE_BIBTEX_FILE
    # checks for absolute path
    assert entry.data['file'] == path.abspath(EXAMPLE_BIBTEX_FILE)


def test_entry_matches():
    """Test match filter."""
    CONFIG.set_config()
    entry = parser.Entry('article', EXAMPLE_ENTRY_DICT)
    # author must match
    _filter = {('author', True): ['Cao']}
    assert entry.matches(_filter, _or=False)
    # author must NOT match
    _filter = {('author', False): ['Coa']}
    assert entry.matches(_filter, _or=False)
    # author and year must match
    _filter = {('author', True): ['Cao'], ('year', True): ['2019']}
    assert entry.matches(_filter, _or=False)
    # author OR year must match
    _filter = {('author', True): ['Cao'], ('year', True): ['2020']}
    assert entry.matches(_filter, _or=True)
    _filter = {('author', True): ['Coa'], ('year', True): ['2019']}
    assert entry.matches(_filter, _or=True)
    # author must NOT match but year must match
    _filter = {('author', False): ['Coa'], ('year', True): ['2019']}
    assert entry.matches(_filter, _or=False)


def test_match_with_wrong_key():
    """Asserts issue #1 is fixed.

    When matches() is called with a key in the filter which does not exist in the entry, the key
    should be ignored and the function should return normally.
    """
    CONFIG.set_config()
    entry = parser.Entry('article', EXAMPLE_ENTRY_DICT)
    _filter = {('tags', False): ['test']}
    assert entry.matches(_filter, _or=False)


def test_to_bibtex():
    """Test to bibtex conversion."""
    pytest.skip("Testing this string is too ambigious. Assumed to be tested by bibtexparser.")


def test_to_yaml():
    """Test to yaml conversion."""
    CONFIG.set_config()
    entry = parser.Entry(EXAMPLE_ENTRY_DICT['ID'], EXAMPLE_ENTRY_DICT)
    yaml_str = entry.to_yaml()
    with open(EXAMPLE_YAML_FILE, 'r') as file:
        assert yaml_str == file.read()


@pytest.mark.parametrize('month_type', ['int', 'str'])
def test_parser_from_bibtex_as_str(month_type):
    """Test parsing a bibtex string.

    Args:
        month_type (str): type to use for storing the 'month' field.
    """
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    CONFIG.config['FORMAT']['month'] = month_type
    reference = EXAMPLE_ENTRY_DICT.copy()
    if month_type == 'str':
        reference['month'] = 'aug'
    with open(EXAMPLE_BIBTEX_FILE, 'r') as file:
        bibtex_str = file.read()
    entries = parser.Entry.from_bibtex(bibtex_str, string=True)
    entry = list(entries.values())[0]
    assert entry.data == reference


@pytest.mark.parametrize('month_type', ['int', 'str'])
def test_parser_from_bibtex_as_file(month_type):
    """Test parsing a bibtex file.

    Args:
        month_type (str): type to use for storing the 'month' field.
    """
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    CONFIG.config['FORMAT']['month'] = month_type
    reference = EXAMPLE_ENTRY_DICT.copy()
    if month_type == 'str':
        reference['month'] = 'aug'
    with open(EXAMPLE_BIBTEX_FILE, 'r') as bibtex_file:
        entries = parser.Entry.from_bibtex(bibtex_file, string=False)
        entry = list(entries.values())[0]
        assert entry.data == reference


@pytest.mark.parametrize('month_type', ['int', 'str'])
def test_parser_from_yaml_as_file(month_type):
    """Test parsing a yaml file.

    Args:
        month_type (str): type to use for storing the 'month' field.
    """
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    CONFIG.config['FORMAT']['month'] = month_type
    reference = EXAMPLE_ENTRY_DICT.copy()
    if month_type == 'str':
        reference['month'] = 'aug'
    # with open(EXAMPLE_YAML_FILE, 'r') as yaml_file:
    entries = parser.Entry.from_yaml(EXAMPLE_YAML_FILE)
    entry = list(entries.values())[0]
    assert entry.data == reference


@pytest.mark.parametrize('month_type', ['int', 'str'])
def test_parser_from_doi(month_type):
    """Test parsing from doi.

    Args:
        month_type (str): type to use for storing the 'month' field.
    """
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    CONFIG.config['FORMAT']['month'] = month_type
    reference = EXAMPLE_ENTRY_DICT.copy()
    if month_type == 'str':
        reference['month'] = 'aug'
    # In this specific case the bib file provided by this DOI includes additional (yet unnecessary)
    # brackets in the escaped special characters of the author field. Thus, we correct for this
    # inconsistency manually before asserting the equality.
    reference['author'] = reference['author'].replace("'a", "'{a}")
    entries = parser.Entry.from_doi('10.1021/acs.chemrev.8b00803')
    if entries == {}:
        pytest.skip("The requests library experienced a ReadTimeout!")
    entry = list(entries.values())[0]
    assert entry.data == reference


def test_parser_from_doi_invalid():
    """Test parsing an invalid DOI."""
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    entries = parser.Entry.from_doi('1812.09976')
    assert not entries
    assert entries == {}


def test_parser_from_isbn():
    """Test parsing from ISBN."""
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    entries = parser.Entry.from_isbn('978-1-449-35573-9')
    entry = list(entries.values())[0]
    assert entry.label == 'Lutz2013'
    assert entry.data['author'] == 'Mark Lutz'
    assert entry.data['pages'] == '1540'
    assert entry.data['title'] == 'Learning Python'
    assert entry.data['year'] == '2013'


# regression test for https://gitlab.com/mrossinek/cobib/-/issues/53
def test_parser_from_isbn_empty():
    """Test parsing an empty ISBN."""
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    entries = parser.Entry.from_isbn('3860704443')
    assert not entries
    assert entries == {}


def test_parser_from_arxiv():
    """Test parsing from arxiv."""
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    reference = EXAMPLE_ENTRY_DICT.copy()
    entries = parser.Entry.from_arxiv('1812.09976')
    entry = list(entries.values())[0]
    assert entry.label == 'Cao2018'
    assert entry.data['archivePrefix'] == 'arXiv'
    assert entry.data['arxivid'].startswith('1812.09976')
    assert entry.data['author'] == reference['author']
    assert entry.data['title'] == reference['title']
    assert entry.data['year'] == '2018'


# regression test for https://gitlab.com/mrossinek/cobib/-/issues/57
def test_parser_from_arxiv_invalid():
    """Test parsing an invalid arXiv ID."""
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    entries = parser.Entry.from_arxiv('10.1021/acs.chemrev.8b00803')
    assert not entries
    assert entries == {}


@pytest.mark.parametrize('month_type', ['int', 'str'])
def test_escape_special_chars(month_type):
    """Test escaping special characters.

    Args:
        month_type (str): type to use for storing the 'month' field.
    """
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    CONFIG.config['FORMAT']['month'] = month_type
    reference = EXAMPLE_ENTRY_DICT.copy()
    if month_type == 'str':
        reference['month'] = 'aug'
    with open('test/example_entry_unescaped.bib', 'r') as bibtex_file:
        entries = parser.Entry.from_bibtex(bibtex_file, string=False)
        entry = list(entries.values())[0]
        assert entry.data == reference


def test_unchanged_umlaut_in_label():
    """Test unchanged Umlaut in labels."""
    root = path.abspath(path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    reference = {
        'ENTRYTYPE': 'book',
        'ID': 'LaTeX_Einführung',
        'title': 'LaTeX Einf{\\"u}hrung',
    }
    with open('test/example_entry_umlaut.bib', 'r') as bibtex_file:
        entries = parser.Entry.from_bibtex(bibtex_file, string=False)
        entry = list(entries.values())[0]
        assert entry.data == reference



#!/usr/bin/env python3
"""CoBib main body."""

import argparse
import inspect
import logging
import logging.config
import sys

from cobib import commands, zsh_helper
from cobib import __version__
from cobib.config import CONFIG
from cobib.database import read_database
from cobib.logging import log_to_stream, log_to_file
from cobib.tui import tui

LOGGER = logging.getLogger(__name__)


def main():
    """Main executable.

    CoBib's main function used to parse optional keyword arguments and subcommands.
    """
    if len(sys.argv) > 1 and any([a[0] == '_' for a in sys.argv]):
        # zsh helper function called
        zsh_main()
        sys.exit()

    # initialize logging
    log_to_stream()

    subcommands = [cmd.split(':')[0] for cmd in zsh_helper.list_commands()]
    parser = argparse.ArgumentParser(prog='CoBib', description="""
                                     Cobib input arguments.
                                     If no arguments are given, the TUI will start as a default.
                                     """)
    parser.add_argument("--version", action="version",
                        version="%(prog)s v{}".format(__version__))
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument("-l", "--logfile", type=argparse.FileType('w'),
                        help="Alternative log file")
    parser.add_argument("-c", "--config", type=argparse.FileType('r'),
                        help="Alternative config file")
    parser.add_argument('command', help="subcommand to be called", choices=subcommands, nargs='?')
    parser.add_argument('args', nargs=argparse.REMAINDER)

    args = parser.parse_args()

    if args.logfile:
        LOGGER.info('Switching to FileHandler logger in %s', args.logfile.name)
        log_to_file('DEBUG' if args.verbose > 1 else 'INFO', logfile=args.logfile.name)

    # set logging verbosity level
    if args.verbose == 1:
        logging.getLogger().setLevel(logging.INFO)
        LOGGER.info('Logging level set to INFO.')
    elif args.verbose > 1:
        logging.getLogger().setLevel(logging.DEBUG)
        LOGGER.info('Logging level set to DEBUG.')

    CONFIG.set_config(args.config)
    try:
        CONFIG.validate()
    except RuntimeError as exc:
        LOGGER.error(exc)
        sys.exit(1)

    if args.command == 'init':
        # the database file may not exist yet, thus we ensure to execute the command before trying
        # to read the database file
        subcmd = getattr(commands, 'InitCommand')()
        subcmd.execute(args.args)
        return

    read_database()
    if not args.command:
        if args.logfile is None:
            LOGGER.info('Switching to FileHandler logger in %s', '/tmp/cobib.log')
            log_to_file('DEBUG' if args.verbose > 1 else 'INFO')
        else:
            LOGGER.info('Already logging to %s. NOT switching to "/tmp/cobib.log"', args.logfile)
        tui()
    else:
        subcmd = getattr(commands, args.command.title()+'Command')()
        subcmd.execute(args.args)


def zsh_main():
    """ZSH helper.

    Main function used by the ZSH completion script.
    """
    helper_avail = ['_'+m[0] for m in inspect.getmembers(zsh_helper) if inspect.isfunction(m[1])]
    parser = argparse.ArgumentParser(description="Process ZSH helper call")
    parser.add_argument('helper', help="zsh helper to be called", choices=helper_avail)
    parser.add_argument('args', nargs=argparse.REMAINDER)

    args = parser.parse_args()

    helper = getattr(zsh_helper, args.helper.strip('_'))
    # any zsh helper function will return a list of the requested items
    for item in helper(args=args.args):
        print(item)


if __name__ == '__main__':
    main()



"""CoBib commands."""

from .add import AddCommand
from .delete import DeleteCommand
from .edit import EditCommand
from .export import ExportCommand
from .init import InitCommand
from .list import ListCommand
from .modify import ModifyCommand
from .open import OpenCommand
from .redo import RedoCommand
from .search import SearchCommand
from .show import ShowCommand
from .undo import UndoCommand


__all__ = [
    "AddCommand",
    "DeleteCommand",
    "EditCommand",
    "ExportCommand",
    "InitCommand",
    "ListCommand",
    "ModifyCommand",
    "OpenCommand",
    "RedoCommand",
    "SearchCommand",
    "ShowCommand",
    "UndoCommand",
    ]



"""Tests for CoBib's config validation."""
# pylint: disable=unused-argument, redefined-outer-name

import os
from pathlib import Path

import pytest
from cobib.config import CONFIG


@pytest.fixture
def setup():
    """Setup."""
    # ensure configuration is empty
    CONFIG.config = {}
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))


def test_base_config(setup):
    """Test the initial configuration passes all validation checks."""
    CONFIG.validate()


@pytest.mark.parametrize(['section'], [
        ['DATABASE'],
        ['FORMAT'],
        ['TUI'],
        ['KEY_BINDINGS'],
        ['COLORS'],
    ])
def test_missing_section(setup, section):
    """Test raised RuntimeError for missing configuration section."""
    with pytest.raises(RuntimeError) as exc_info:
        del CONFIG.config[section]
        CONFIG.validate()
    assert section in str(exc_info.value)


@pytest.mark.parametrize(['section', 'field'], [
        ['DATABASE', 'file'],
        ['DATABASE', 'open'],
        ['DATABASE', 'grep'],
        ['DATABASE', 'search_ignore_case'],
        ['FORMAT', 'month'],
        ['FORMAT', 'ignore_non_standard_types'],
        ['FORMAT', 'default_entry_type'],
        ['TUI', 'default_list_args'],
        ['TUI', 'prompt_before_quit'],
        ['TUI', 'reverse_order'],
        ['TUI', 'scroll_offset'],
        ['COLORS', 'cursor_line_fg'],
        ['COLORS', 'cursor_line_bg'],
        ['COLORS', 'top_statusbar_fg'],
        ['COLORS', 'top_statusbar_bg'],
        ['COLORS', 'bottom_statusbar_fg'],
        ['COLORS', 'bottom_statusbar_bg'],
        ['COLORS', 'search_label_fg'],
        ['COLORS', 'search_label_bg'],
        ['COLORS', 'search_query_fg'],
        ['COLORS', 'search_query_bg'],
        ['COLORS', 'popup_help_fg'],
        ['COLORS', 'popup_help_bg'],
        ['COLORS', 'popup_stdout_fg'],
        ['COLORS', 'popup_stdout_bg'],
        ['COLORS', 'popup_stderr_fg'],
        ['COLORS', 'popup_stderr_bg'],
        ['COLORS', 'selection_fg'],
        ['COLORS', 'selection_bg'],
    ])
def test_database_section(setup, section, field):
    """Test raised RuntimeError for missing config fields."""
    with pytest.raises(RuntimeError) as exc_info:
        del CONFIG.config.get(section, {})[field]
        CONFIG.validate()
    assert f'{section}/{field}' in str(exc_info.value)


@pytest.mark.parametrize(['color'], [
        ['cursor_line_fg'],
        ['cursor_line_bg'],
        ['top_statusbar_fg'],
        ['top_statusbar_bg'],
        ['bottom_statusbar_fg'],
        ['bottom_statusbar_bg'],
        ['search_label_fg'],
        ['search_label_bg'],
        ['search_query_fg'],
        ['search_query_bg'],
        ['popup_help_fg'],
        ['popup_help_bg'],
        ['popup_stdout_fg'],
        ['popup_stdout_bg'],
        ['popup_stderr_fg'],
        ['popup_stderr_bg'],
        ['selection_fg'],
        ['selection_bg'],
    ])
def test_valid_tui_colors(setup, color):
    """Test curses color specification validation."""
    with pytest.raises(RuntimeError) as exc_info:
        CONFIG.config.get('COLORS', {})[color] = 'test'
        CONFIG.validate()
    assert str(exc_info.value) == 'Unknown color specification: test'



"""CoBib export command."""

import argparse
import logging
import os
import sys
from zipfile import ZipFile

from cobib.config import CONFIG
from .base_command import ArgumentParser, Command
from .list import ListCommand

LOGGER = logging.getLogger(__name__)


class ExportCommand(Command):
    """Export Command."""

    name = 'export'

    def execute(self, args, out=sys.stdout):
        """Export database.

        Exports all entries matched by the filter queries (see the list docs).
        Currently supported exporting formats are:
        * BibLaTex databases
        * zip archives

        Args: See base class.
        """
        LOGGER.debug('Starting Export command.')
        parser = ArgumentParser(prog="export", description="Export subcommand parser.")
        parser.add_argument("-b", "--bibtex", type=argparse.FileType('a'),
                            help="BibLaTeX output file")
        parser.add_argument("-z", "--zip", type=argparse.FileType('a'),
                            help="zip output file")
        parser.add_argument("-s", "--selection", action="store_true",
                            help="When specified, the `filter` argument will be interpreted as "
                            "a list of entry labels rather than arguments for the `list` command.")
        parser.add_argument('filter', nargs='*',
                            help="You can specify filters as used by the `list` command in order "
                            "to select a subset of labels to be modified. To ensure this works as "
                            "expected you should add the pseudo-argument '--' before the list of "
                            "filters. See also `list --help` for more information.")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_intermixed_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        if largs.bibtex is None and largs.zip is None:
            msg = "No output file specified!"
            print("Error: " + msg, file=sys.stderr)
            LOGGER.error(msg)
            return
        if largs.zip is not None:
            largs.zip = ZipFile(largs.zip.name, 'w')
        out = open(os.devnull, 'w')
        if largs.selection:
            LOGGER.info('Selection given. Interpreting `filter` as a list of labels')
            labels = largs.filter
        else:
            LOGGER.debug('Gathering filtered list of entries to be exported.')
            labels = ListCommand().execute(largs.filter, out=out)

        try:
            for label in labels:
                LOGGER.debug('Exporting entry "%s".', label)
                entry = CONFIG.config['BIB_DATA'][label]
                if largs.bibtex is not None:
                    entry_str = entry.to_bibtex()
                    largs.bibtex.write(entry_str)
                if largs.zip is not None:
                    if 'file' in entry.data.keys() and entry.data['file'] is not None:
                        LOGGER.debug('Adding "%s" associated with "%s" to the zip file.',
                                     entry.data['file'], label)
                        largs.zip.write(entry.data['file'], label+'.pdf')
        except KeyError:
            print("Error: No entry with the label '{}' could be found.".format(largs.label))

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Export command triggered from TUI.')
        # handle input via prompt
        if tui.selection:
            tui.execute_command('export -s', pass_selection=True)
        else:
            tui.execute_command('export')



"""CoBib configuration module."""

from copy import deepcopy
import configparser
import io
import logging
import os
import sys

LOGGER = logging.getLogger(__name__)

ANSI_COLORS = [
    'black',
    'red',
    'green',
    'yellow',
    'blue',
    'magenta',
    'cyan',
    'white',
]

DEFAULTS = {
    'DATABASE': {
        'file': os.path.expanduser('~/.local/share/cobib/literature.yaml'),
        'git': False,
        'open': 'xdg-open' if sys.platform.lower() == 'linux' else 'open',
        'grep': 'grep',
        'search_ignore_case': False,
    },
    'FORMAT': {
        'month': 'int',
        'ignore_non_standard_types': False,
        'default_entry_type': 'article',
    },
    'TUI': {
        'default_list_args': '-l',
        'prompt_before_quit': True,
        'reverse_order': True,
        'scroll_offset': 3,
    },
    'KEY_BINDINGS': {
    },
    'COLORS': {
        'cursor_line_fg': 'white',
        'cursor_line_bg': 'cyan',
        'top_statusbar_fg': 'black',
        'top_statusbar_bg': 'yellow',
        'bottom_statusbar_fg': 'black',
        'bottom_statusbar_bg': 'yellow',
        'search_label_fg': 'blue',
        'search_label_bg': 'black',
        'search_query_fg': 'red',
        'search_query_bg': 'black',
        'popup_help_fg': 'white',
        'popup_help_bg': 'green',
        'popup_stdout_fg': 'white',
        'popup_stdout_bg': 'blue',
        'popup_stderr_fg': 'white',
        'popup_stderr_bg': 'red',
        'selection_fg': 'white',
        'selection_bg': 'magenta',
    },
}

XDG_CONFIG_FILE = '~/.config/cobib/config.ini'


class Config:
    """Class used solely for the global configuration object."""

    def __init__(self):
        """Initializes the configuration data dictionary."""
        self.config = {}

    def set_config(self, configpath=None):
        """Sets the configuration.

        If a configuration file is provided as a keyword argument it is used instead of the default
        paths. The configparser module is used to parse the INI configuration file.

        Args:
            configpath (str or io.TextIOWrapper, optional): the path to an optional configuration
                                                            file.
        """
        ini_conf = configparser.ConfigParser()
        ini_conf.optionxform = str  # makes option names case-sensitive!
        # load default configuration
        ini_conf.read_dict(DEFAULTS)

        # read ini config file
        if configpath is not None:
            if isinstance(configpath, io.TextIOWrapper):
                configpath = configpath.name
            LOGGER.info('Loading configuration from %s', configpath)
            ini_conf.read(configpath)
        elif os.path.exists(os.path.expanduser(XDG_CONFIG_FILE)):
            LOGGER.info('Loading configuration from default location: %s',
                        os.path.expanduser(XDG_CONFIG_FILE))
            ini_conf.read(os.path.expanduser(XDG_CONFIG_FILE))

        # overwrite settings
        for section in ini_conf.sections():
            self.config[section] = deepcopy(ini_conf[section])

    def validate(self):
        """Validates the configuration at runtime."""
        LOGGER.info('Validating the runtime configuration.')

        # DATABASE section
        LOGGER.debug('Validing the DATABASE configuration section.')
        self._assert(self.config.get('DATABASE', None) is not None,
                     "Missing DATABASE section.")
        self._assert(isinstance(self.config.get('DATABASE', {}).get('file', None), str),
                     "DATABASE/file should be a string.")
        self._assert(isinstance(self.config.get('DATABASE', {}).getboolean('git', None), bool),
                     "DATABASE/git should be a boolean.")
        self._assert(isinstance(self.config.get('DATABASE', {}).get('open', None), str),
                     "DATABASE/open should be a string.")
        self._assert(isinstance(self.config.get('DATABASE', {}).get('grep', None), str),
                     "DATABASE/grep should be a string.")
        self._assert(isinstance(
            self.config.get('DATABASE', {}).getboolean('search_ignore_case', None), bool),
                     "DATABASE/search_ignore_case should be a boolean.")

        # FORMAT section
        LOGGER.debug('Validing the FORMAT configuration section.')
        self._assert(self.config.get('FORMAT', None) is not None,
                     "Missing FORMAT section.")
        self._assert(self.config.get('FORMAT', {}).get('month', None) in ('int', 'str'),
                     "FORMAT/month should be either 'int' or 'str'.")
        self._assert(isinstance(
            self.config.get('FORMAT', {}).getboolean('ignore_non_standard_types', None), bool),
                     "FORMAT/ignore_non_standard_types should be a boolean.")
        self._assert(isinstance(self.config.get('FORMAT', {}).get('default_entry_type', None), str),
                     "FORMAT/default_entry_type should be a string.")

        # TUI section
        LOGGER.debug('Validing the TUI configuration section.')
        self._assert(self.config.get('TUI', None) is not None,
                     "Missing TUI section.")
        self._assert(isinstance(self.config.get('TUI', {}).get('default_list_args', None), str),
                     "TUI/default_list_args should be a string.")
        self._assert(isinstance(
            self.config.get('TUI', {}).getboolean('prompt_before_quit', None), bool),
                     "TUI/prompt_before_quit should be a boolean.")
        self._assert(isinstance(
            self.config.get('TUI', {}).getboolean('reverse_order', None), bool),
                     "TUI/reverse_order should be a boolean.")
        self._assert(isinstance(
            self.config.get('TUI', {}).getint('scroll_offset', None), int),
                     "TUI/scroll_offset should be an integer.")

        # KEY_BINDINGS section
        LOGGER.debug('Validing the KEY_BINDINGS configuration section.')
        self._assert(self.config.get('KEY_BINDINGS', None) is not None,
                     "Missing KEY_BINDINGS section.")
        # actual key bindings are asserted when mapped by the TUI instance

        # COLORS section
        LOGGER.debug('Validing the COLORS configuration section.')
        self._assert(self.config.get('COLORS', None) is not None,
                     "Missing COLORS section.")
        for name in DEFAULTS['COLORS']:
            self._assert(name in self.config.get('COLORS', {}).keys(),
                         f"Missing value for COLORS/{name}")
        available_colors = ('black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white')
        for name, color in self.config.get('COLORS', {}).items():
            if name not in DEFAULTS['COLORS'] and name not in available_colors:
                LOGGER.warning('Ignoring unknown TUI color: %s', name)
            self._assert(color in available_colors or
                         (len(color.strip('#')) == 6 and
                          tuple(int(color.strip('#')[i:i+2], 16) for i in (0, 2, 4))),
                         f"Unknown color specification: {color}")

    @staticmethod
    def _assert(expression, error):
        """Asserts the expression is True.

        Raises:
            RuntimeError with the specified error string.
        """
        if not expression:
            raise RuntimeError(error)

    def get_ansi_color(self, name):
        """Returns an ANSI color code for the named color.

        Args:
            name (str): a named color as specified in the configuration *excluding* the `_fg` or
                        `_bg` suffix.

        Returns:
            A string representing the foreground and background ANSI color code.
        """
        fg_color = 30 + ANSI_COLORS.index(self.config['COLORS'].get(name + '_fg'))
        bg_color = 40 + ANSI_COLORS.index(self.config['COLORS'].get(name + '_bg'))

        return f'\x1b[{fg_color};{bg_color}m'


CONFIG = Config()



"""CoBib TUI module."""

import curses

from .buffer import TextBuffer, InputBuffer
from .tui import TUI


__all__ = ["TUI",
           "TextBuffer",
           "InputBuffer"]


def tui():
    """Main executable for the curses-TUI."""
    curses.wrapper(TUI)



"""Tests for CoBib's TUI."""
# pylint: disable=unused-argument, redefined-outer-name

import curses
import fcntl
import os
import select
import termios
from array import array
from pathlib import Path

import pyte
import pytest
from cobib import __version__ as version
from cobib.commands import AddCommand, DeleteCommand
from cobib.config import CONFIG
from cobib.database import read_database
from cobib.tui import TextBuffer, TUI


@pytest.fixture
def setup():
    """Setup."""
    # ensure configuration is empty
    CONFIG.config = {}
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    # NOTE: normally you would never trigger an Add command before reading the database but in this
    # controlled testing scenario we can be certain that this is fine
    AddCommand().execute(['-b', './test/dummy_scrolling_entry.bib'])
    read_database()
    yield setup
    DeleteCommand().execute(['dummy_entry_for_scroll_testing'])


def assert_list_view(screen, current, expected):
    """Asserts the list view of the TUI."""
    term_width = len(screen.buffer[0])
    # assert default colors
    assert [c.bg for c in screen.buffer[0].values()] == ['brown'] * term_width
    assert [c.bg for c in screen.buffer[len(screen.buffer)-2].values()] == ['brown'] * term_width
    # the top statusline contains the version info and number of entries
    assert f"CoBib v{version} - {len(expected)} Entries" in screen.display[0]
    # check current line
    if current >= 0:
        assert [c.fg for c in screen.buffer[current].values()] == ['white'] * term_width
        assert [c.bg for c in screen.buffer[current].values()] == ['cyan'] * term_width
    # the entries per line
    for idx, label in enumerate(expected):
        # offset of 1 due to top statusline
        assert label in screen.display[idx+1]
    # the bottom statusline should contain at least parts of the information string
    assert screen.display[-2].strip() in TUI.infoline()
    # the prompt line should be empty
    assert screen.display[-1].strip() == ""


def assert_help_screen(screen, only_title=False):
    """Asserts the contents of the Help screen."""
    header_check = ["CoBib TUI Help" in line for line in screen.display]
    assert any(header_check)
    if not only_title:
        offset = header_check.index(True)
        for cmd, desc in TUI.HELP_DICT.items():
            assert any("{:<8} {}".format(cmd+':', desc) in line for line in
                       screen.display[2+offset:19+offset])


def assert_no_help_window_artefacts(screen):
    """Asserts issue #20 remains fixed."""
    assert_list_view(screen, 1, [
        'dummy_entry_for_scroll_testing', 'knuthwebsite', 'latexcompanion', 'einstein'
    ])
    for line in range(5, 22):
        assert screen.display[line].strip() == ''


def assert_scroll(screen, update, direction):
    """Asserts cursor-line position after scrolling.

    Attention: The values of update *strongly* depend on the contents of the dummy scrolling entry.
    """
    term_width = len(screen.buffer[0])
    if direction == 'y' or update == 0:
        assert [c.fg for c in screen.buffer[1 + update].values()] == ['white'] * term_width
        assert [c.bg for c in screen.buffer[1 + update].values()] == ['cyan'] * term_width
    elif direction == 'x':
        # TODO actually use the update information
        assert [c.fg for c in screen.buffer[1].values()] == ['white'] * term_width
        assert [c.bg for c in screen.buffer[1].values()] == ['cyan'] * term_width


def assert_wrap(screen, state):
    """Asserts the viewport buffer is wrapped."""
    if state:
        assert screen.display[2][:2] == TextBuffer.INDENT + ' '
    else:
        assert screen.display[2][:2] == 'kn'


def assert_show(screen):
    """Asserts the show menu."""
    with open('./test/dummy_scrolling_entry.bib', 'r') as source:
        for screen_line, source_line in zip(screen.display[1:5], source.readlines()):
            assert screen_line.strip() in source_line.strip()


def assert_add(screen):
    """Asserts the add prompt."""
    try:
        assert f"CoBib v{version} - 5 Entries" in screen.display[0]
        assert "Cao_2019" in screen.display[1]
        assert "dummy_entry_for_scroll_testing" in screen.display[2]
        assert "knuthwebsite" in screen.display[3]
        assert "latexcompanion" in screen.display[4]
        assert "einstein" in screen.display[5]
    finally:
        DeleteCommand().execute(['Cao_2019'])


def assert_delete(screen):
    """Asserts entry is deleted.

    This also ensures it is added again after successful deletion.
    """
    try:
        assert f"CoBib v{version} - 3 Entries" in screen.display[0]
        assert not any("dummy_entry_for_scroll_testing" in line for line in screen.display[4:21])
    finally:
        AddCommand().execute(['-b', './test/dummy_scrolling_entry.bib'])


def assert_editor(screen):
    """Asserts the editor opens."""
    start = 0
    while '---' not in screen.display[start]:
        start += 1
    end = start+1
    while '...' not in screen.display[end]:
        end += 1
    with open('./test/example_literature.yaml', 'r') as source:
        for screen_line, source_line in zip(screen.display[start:end], source.readlines()):
            if screen_line.strip() == '~':
                # vim populates lines after the end of the file with single '~' characters
                break
            # assert the source line is fully part of the line visible in the editor window
            assert source_line.strip() in screen_line.strip()


def assert_prompt(screen, contents):
    """Asserts the prompt line contents."""
    assert screen.display[-1].strip() == contents


def assert_search_view(screen, label, search, selected):
    """Asserts the search screen contents."""
    term_width = len(screen.buffer[0])
    if selected:
        assert [c.bg for c in screen.buffer[1].values()] == \
            ['magenta'] * len(label) + \
            ['default'] * (term_width - len(label))
    else:
        assert [c.fg for c in screen.buffer[1].values()] == \
            ['blue'] * len(label) + \
            ['default'] * (term_width - len(label))
    offset = screen.display[2].index(search)
    assert [c.fg for c in screen.buffer[2].values()] == \
        ['default'] * offset + \
        ['red'] * len(search) + \
        ['default'] * (term_width - len(search) - offset)


def assert_select_list_view(screen, current, selected, labels):
    """Asserts the selection in the list view of the TUI."""
    term_width = len(screen.buffer[0])
    for sel, lab in zip(selected, labels):
        assert [c.bg for c in screen.buffer[sel].values()] == ['magenta'] * len(lab) + \
                ['cyan' if sel == current else 'default'] * (term_width - len(lab))


def assert_select_show_view(screen, current):
    """Asserts the selection in the show view."""
    label_len = len('dummy_entry_for_scroll_testing')
    with open('./test/dummy_scrolling_entry.bib', 'r') as source:
        for idx, (screen_line, source_line) in enumerate(zip(screen.display[1:5],
                                                             source.readlines())):
            if idx == 0:
                assert [c.bg for c in screen.buffer[1].values()] == \
                    ['cyan' if current else 'default'] * 6 + \
                    ['magenta'] * label_len + \
                    ['cyan' if current else 'default'] * (len(screen.buffer[0]) - label_len - 6)
            assert screen_line.strip() in source_line.strip()


@pytest.mark.parametrize(['keys', 'assertion', 'assertion_kwargs'], [
        ['', assert_list_view, {
            'current': 1, 'expected': [
                'dummy_entry_for_scroll_testing', 'knuthwebsite', 'latexcompanion', 'einstein'
            ]}],
        ['?q', assert_no_help_window_artefacts, {}],
        ['?', assert_help_screen, {}],
        ['w', assert_wrap, {'state': True}],
        ['ww', assert_wrap, {'state': False}],
        ['\n', assert_show, {}],
        ['a-b ./test/example_entry.bib\n', assert_add, {}],
        ['d', assert_delete, {}],
        ['Ge', assert_editor, {}],
        ['x', assert_prompt, {'contents': ":export"}],
        [':', assert_prompt, {'contents': ":"}],
        [':' + 100 * 'j', assert_prompt, {'contents': 'j' * 79}],  # regression-test on #48
        ['f++ID einstein\n', assert_list_view, {
            'current': 1, 'expected': ['einstein']}],
        ['f--ID einstein\n', assert_list_view, {
            'current': 1, 'expected': [
                'dummy_entry_for_scroll_testing', 'knuthwebsite', 'latexcompanion'
            ]}],
        ['f++ID einstein ++ID knuthwebsite\n', assert_list_view, {
            'current': -1, 'expected': []}],
        ['f++ID einstein ++ID knuthwebsite -x\n', assert_list_view, {
            'current': 1, 'expected': ['knuthwebsite', 'einstein']}],
        ['syear\n', assert_list_view, {
            'current': 1, 'expected': [
                'latexcompanion', 'knuthwebsite', 'einstein', 'dummy_entry_for_scroll_testing'
            ]}],
        ['/einstein\njj', assert_search_view, {
            'label': 'einstein',
            'search': 'einstein',
            'selected': False
        }],
        ['/einstein\njjv', assert_search_view, {
            'label': 'einstein',
            'search': 'einstein',
            'selected': True
        }],
        ['Gv/einstein\njj', assert_search_view, {
            'label': 'einstein',
            'search': 'einstein',
            'selected': True
        }],
        ['/einstein\nvq', assert_select_list_view, {'current': 1, 'selected': [4],
                                                    'labels': ['einstein']}],
        ['v', assert_select_list_view, {'current': 1, 'selected': [1],
                                        'labels': ['dummy_entry_for_scroll_testing']}],
        ['vj', assert_select_list_view, {'current': 2, 'selected': [1],
                                         'labels': ['dummy_entry_for_scroll_testing']}],
        ['jvjv', assert_select_list_view, {'current': 3, 'selected': [2, 3],
                                           'labels': ['knuthwebsite', 'latexcompanion']}],
        ['vv', assert_list_view, {  # assert that re-selection toggles
            'current': 1, 'expected': [
                'dummy_entry_for_scroll_testing', 'knuthwebsite', 'latexcompanion', 'einstein'
            ]}],
        ['v\n', assert_select_show_view, {'current': True}],
        ['\nv', assert_select_show_view, {'current': True}],
        ['\nvj', assert_select_show_view, {'current': False}],
        ['\nvq', assert_select_list_view, {'current': 1, 'selected': [1],
                                           'labels': ['dummy_entry_for_scroll_testing']}],
    ])
def test_tui(setup, keys, assertion, assertion_kwargs):
    """Test TUI.

    Args:
        setup: runs pytest fixture.
        keys (str): keys to be send to the CoBib TUI.
        assertion (Callable): function to run the assertions for the key to be tested.
        assertion_kwargs (dict): additional keyword arguments for assertion function.
    """
    # create pseudo-terminal
    pid, f_d = os.forkpty()
    if pid == 0:
        # child process spawns TUI
        curses.wrapper(TUI)
    else:
        # parent process sets up virtual screen of identical size
        screen = pyte.Screen(80, 24)
        stream = pyte.ByteStream(screen)
        # send keys char-wise to TUI
        for key in keys:
            os.write(f_d, str.encode(key))
        # scrape pseudo-terminal's screen
        while True:
            try:
                [f_d], _, _ = select.select([f_d], [], [], 1)
            except (KeyboardInterrupt, ValueError):
                # either test was interrupted or file descriptor of child process provides nothing
                # to be read
                break
            else:
                try:
                    # scrape screen of child process
                    data = os.read(f_d, 1024)
                    stream.feed(data)
                except OSError:
                    # reading empty
                    break
        for line in screen.display:
            print(line)
        assertion(screen, **assertion_kwargs)


def assert_config_color(screen, colors):
    """Assert configured colors."""
    term_width = len(screen.buffer[0])
    assert [c.bg for c in screen.buffer[0].values()] == [colors['top_statusbar_bg']] * term_width
    assert [c.fg for c in screen.buffer[0].values()] == [colors['top_statusbar_fg']] * term_width
    assert [c.bg for c in screen.buffer[len(screen.buffer)-2].values()] == \
        [colors['bottom_statusbar_bg']] * term_width
    assert [c.fg for c in screen.buffer[len(screen.buffer)-2].values()] == \
        [colors['bottom_statusbar_fg']] * term_width
    assert [c.bg for c in screen.buffer[1].values()] == [colors['cursor_line_bg']] * term_width
    assert [c.fg for c in screen.buffer[1].values()] == [colors['cursor_line_fg']] * term_width


def test_tui_config_color():
    """Test TUI color configuration."""
    # ensure configuration is empty
    CONFIG.config = {}
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    # overwrite color configuration
    CONFIG.config['COLORS']['top_statusbar_bg'] = 'red'
    CONFIG.config['COLORS']['top_statusbar_fg'] = 'blue'
    CONFIG.config['COLORS']['bottom_statusbar_bg'] = 'green'
    CONFIG.config['COLORS']['bottom_statusbar_fg'] = 'magenta'
    CONFIG.config['COLORS']['cursor_line_bg'] = 'white'
    CONFIG.config['COLORS']['cursor_line_fg'] = 'black'
    read_database()
    test_tui(None, '', assert_config_color, {'colors': CONFIG.config['COLORS']})


@pytest.mark.parametrize(['command', 'key'], [
        ['Show', 'p'],  # previously unused key
        ['Show', 'o'],  # should overwrite previously used key with other command
    ])
def test_tui_config_keys(command, key):
    """Test TUI key binding configuration."""
    # ensure configuration is empty
    CONFIG.config = {}
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    # overwrite key binding configuration
    CONFIG.config['KEY_BINDINGS'][command] = key
    # NOTE: normally you would never trigger an Add command before reading the database but in this
    # controlled testing scenario we can be certain that this is fine
    AddCommand().execute(['-b', './test/dummy_scrolling_entry.bib'])
    read_database()
    try:
        test_tui(None, key, assert_show, {})
    finally:
        DeleteCommand().execute(['dummy_entry_for_scroll_testing'])


def assert_quit(screen, prompt):
    """Asserts the quit prompt."""
    if prompt == 'True':
        assert screen.display[-1].strip() == 'Do you really want to quit CoBib? [y/n]'
    elif prompt == 'False':
        assert screen.display[-1].strip() == ''
    else:
        assert not 'Unexpected prompt setting!'


@pytest.mark.parametrize(['setting', 'keys'], [
        ['True', 'q'],
        ['False', 'q'],
    ])
def test_tui_quit_prompt(setting, keys):
    """Test the prompt_before_quit setting of the TUI."""
    # ensure configuration is empty
    CONFIG.config = {}
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    # set prompt_before_quit setting
    CONFIG.config['TUI']['prompt_before_quit'] = setting
    read_database()
    test_tui(None, keys, assert_quit, {'prompt': setting})


def assert_open(screen):
    """Asserts the open menu."""
    assert '1: [file] /tmp/a.txt' in screen.display[16]
    assert '2: [file] /tmp/b.txt' in screen.display[17]
    assert '3: [url] https://www.duckduckgo.com' in screen.display[18]
    assert '4: [url] https://www.google.com' in screen.display[19]
    assert "Entry to open [Type 'help' for more info]:" in screen.display[20]


def test_tui_open_menu():
    """Test the open prompt menu for multiple associated files."""
    # ensure configuration is empty
    CONFIG.config = {}
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    # NOTE: normally you would never trigger an Add command before reading the database but in this
    # controlled testing scenario we can be certain that this is fine
    AddCommand().execute(['-b', './test/dummy_multi_file_entry.bib'])
    read_database()
    try:
        test_tui(None, 'o', assert_open, {})
    finally:
        DeleteCommand().execute(['dummy_multi_file_entry'])


@pytest.mark.parametrize(['keys', 'assertion', 'assertion_kwargs'], [
        ['', assert_list_view, {
            'current': 1, 'expected': [
                'dummy_entry_for_scroll_testing', 'knuthwebsite', 'latexcompanion', 'einstein'
            ]}],
        ['?', assert_help_screen, {'only_title': True}],
    ])
def test_tui_resize(setup, keys, assertion, assertion_kwargs):
    """Test TUI resize handling.

    Includes a regression test against #58 by opening the help popup.

    Args:
        setup: runs pytest fixture.
        keys (str): keys to be send to the CoBib TUI.
        assertion (Callable): function to run the assertions for the key to be tested.
        assertion_kwargs (dict): additional keyword arguments for assertion function.
    """
    # create pseudo-terminal
    pid, f_d = os.forkpty()
    if pid == 0:
        # child process spawns TUI
        curses.wrapper(TUI)
    else:
        # resize pseudo terminal
        fcntl.ioctl(f_d, termios.TIOCSWINSZ, array('h', [10, 45, 1200, 220]))
        # parent process sets up virtual screen of identical size
        screen = pyte.Screen(45, 10)
        stream = pyte.ByteStream(screen)
        # send keys char-wise to TUI
        for key in keys:
            os.write(f_d, str.encode(key))
        # scrape pseudo-terminal's screen
        while True:
            try:
                [f_d], _, _ = select.select([f_d], [], [], 1)
            except (KeyboardInterrupt, ValueError):
                # either test was interrupted or file descriptor of child process provides nothing
                # to be read
                break
            else:
                try:
                    # scrape screen of child process
                    data = os.read(f_d, 1024)
                    stream.feed(data)
                except OSError:
                    # reading empty
                    break
        for line in screen.display:
            print(line)
        assertion(screen, **assertion_kwargs)


@pytest.mark.parametrize(['keys', 'assertion', 'assertion_kwargs'], [
        # vertical scrolling
        ['G', assert_scroll, {'update': 20, 'direction': 'y'}],
        ['Gg', assert_scroll, {'update': 0, 'direction': 'y'}],
        ['j', assert_scroll, {'update': 1, 'direction': 'y'}],
        ['jjk', assert_scroll, {'update': 1, 'direction': 'y'}],
        # assert scrolloff value of `3` is respected
        [''.join(['j'] * 20), assert_scroll, {'update': 17, 'direction': 'y'}],
        [''.join(['j'] * 21), assert_scroll, {'update': 18, 'direction': 'y'}],
        [''.join(['j'] * 22), assert_scroll, {'update': 19, 'direction': 'y'}],
        ['G' + ''.join(['k'] * 20), assert_scroll, {'update': 3, 'direction': 'y'}],
        ['G' + ''.join(['k'] * 21), assert_scroll, {'update': 2, 'direction': 'y'}],
        ['G' + ''.join(['k'] * 22), assert_scroll, {'update': 1, 'direction': 'y'}],
        # horizontal scrolling
        ['l', assert_scroll, {'update': 1, 'direction': 'x'}],
        ['llh', assert_scroll, {'update': 1, 'direction': 'x'}],
        ['$', assert_scroll, {'update': 23, 'direction': 'x'}],
        ['$0', assert_scroll, {'update': 0, 'direction': 'x'}],
    ])
def test_tui_scrolling(keys, assertion, assertion_kwargs):
    """Test TUI scrolling behavior."""
    # ensure configuration is empty
    CONFIG.config = {}
    root = os.path.abspath(os.path.dirname(__file__))
    CONFIG.set_config(Path(root + '/../cobib/docs/debug.ini'))
    # overwrite database file
    CONFIG.config['DATABASE']['file'] = './test/scrolling_database.yaml'
    read_database()
    test_tui(None, keys, assertion, assertion_kwargs)



"""CoBib edit command."""

import argparse
import logging
import os
import sys
import tempfile

from cobib.config import CONFIG
from cobib.database import read_database
from cobib.parser import Entry
from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class EditCommand(Command):
    """Edit Command."""

    name = 'edit'

    def execute(self, args, out=sys.stdout):
        """Edit entry.

        Opens an existing entry for manual editing.

        Args: See base class.
        """
        LOGGER.debug('Starting Edit command.')
        parser = ArgumentParser(prog="edit", description="Edit subcommand parser.")
        parser.add_argument("label", type=str, help="label of the entry")
        parser.add_argument("-a", "--add", action='store_true',
                            help="if specified, will add a new entry for unknown labels")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        try:
            entry = CONFIG.config['BIB_DATA'][largs.label]
            prv = entry.to_yaml()
            if largs.add:
                LOGGER.warning("Entry '%s' already exists! Ignoring the `--add` argument.",
                               largs.label)
                largs.add = False
        except KeyError:
            # No entry for given label found
            if largs.add:
                # add a new entry for the unknown label
                entry = Entry(largs.label,
                              {'ID': largs.label,
                               'ENTRYTYPE': CONFIG.config['FORMAT']['default_entry_type']})
                prv = entry.to_yaml()
            else:
                msg = f"No entry with the label '{largs.label}' could be found.\n" \
                    + "Use `--add` to add a new entry with this label."
                LOGGER.error(msg)
                return

        LOGGER.debug('Creating temporary file.')
        tmp_file = tempfile.NamedTemporaryFile(mode='w+', prefix='cobib-', suffix='.yaml')
        tmp_file.write(prv)
        tmp_file.flush()
        LOGGER.debug('Starting editor "%s".', os.environ['EDITOR'])
        status = os.system(os.environ['EDITOR'] + ' ' + tmp_file.name)
        assert status == 0
        LOGGER.debug('Editor finished successfully.')
        with open(tmp_file.name, 'r') as edited:
            nxt = ''.join(edited.readlines()[1:])
        tmp_file.close()
        assert not os.path.exists(tmp_file.name)
        if prv == nxt:
            LOGGER.info('No changes detected.')
            return
        conf_database = CONFIG.config['DATABASE']
        file = os.path.expanduser(conf_database['file'])
        with open(file, 'r') as bib:
            lines = bib.readlines()
        entry_to_be_replaced = False
        with open(file, 'w') as bib:
            for line in lines:
                if line.startswith(largs.label + ':'):
                    LOGGER.debug('Entry "%s" found. Starting to replace lines.', largs.label)
                    entry_to_be_replaced = True
                    continue
                if entry_to_be_replaced and line.startswith('...'):
                    LOGGER.debug('Reached end of entry "%s".', largs.label)
                    entry_to_be_replaced = False
                    bib.writelines(nxt)
                    continue
                if not entry_to_be_replaced:
                    bib.write(line)
            if largs.add:
                # append new entry
                bib.write('---\n')
                bib.writelines(nxt)
                msg = f"'{largs.label}' was added to the database."
                print(msg)
                LOGGER.info(msg)

        self.git(args=vars(largs))

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Edit command triggered from TUI.')
        # get current label
        label, _ = tui.viewport.get_current_label()
        # populate buffer with entry data
        EditCommand().execute([label])
        # update bibliography data
        read_database()
        # redraw total screen after closing external editor
        LOGGER.debug('Manually redrawing TUI to clear out any editor artefacts.')
        tui.resize_handler(None, None)
        # update database list
        tui.viewport.update_list()



"""CoBib."""
import os
import subprocess

__version__ = '2.6.1'

if os.path.exists(os.path.dirname(os.path.dirname(os.path.abspath(__file__))) + '/.git'):
    # if installed from source, append HEAD commit SHA to version info as metadata
    proc = subprocess.Popen(['git', 'rev-parse', 'HEAD'], stdout=subprocess.PIPE)
    git_revision, _ = proc.communicate()
    __version__ += '+' + git_revision.decode('utf-8')[:7]



"""CoBib search command."""

import argparse
import logging
import os
import re
import shlex
import sys

from cobib import __version__
from cobib.config import CONFIG
from .base_command import ArgumentParser, Command
from .list import ListCommand

LOGGER = logging.getLogger(__name__)


class SearchCommand(Command):
    """Search Command."""

    name = 'search'

    def execute(self, args, out=sys.stdout):
        """Search database.

        Searches the database recursively (i.e. including any associated files) using `grep` for a
        query string.

        Args: See base class.
        """
        LOGGER.debug('Starting Search command.')
        parser = ArgumentParser(prog="search", description="Search subcommand parser.")
        parser.add_argument("query", type=str, help="text to search for")
        parser.add_argument("-c", "--context", type=int, default=1,
                            help="number of context lines to provide for each match")
        parser.add_argument("-i", "--ignore-case", action="store_true",
                            help="ignore case for searching")
        parser.add_argument('filter', nargs='*',
                            help="You can specify filters as used by the `list` command in order "
                            "to select a subset of labels to be modified. To ensure this works as "
                            "expected you should add the pseudo-argument '--' before the list of "
                            "filters. See also `list --help` for more information.")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_intermixed_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return None

        labels = ListCommand().execute(largs.filter, out=open(os.devnull, 'w'))
        LOGGER.debug('Available entries to search: %s', labels)

        ignore_case = CONFIG.config['DATABASE'].getboolean('search_ignore_case', False) or \
            largs.ignore_case
        re_flags = re.IGNORECASE if ignore_case else 0
        LOGGER.debug('The search will be performed case %ssensitive', 'in' if ignore_case else '')

        hits = 0
        output = []
        for label in labels.copy():
            entry = CONFIG.config['BIB_DATA'][label]
            matches = entry.search(largs.query, largs.context, ignore_case)
            if not matches:
                labels.remove(label)
                continue

            hits += len(matches)
            LOGGER.debug('Entry "%s" includes %d hits.', label, hits)
            title = f"{label} - {len(matches)} match" + ("es" if len(matches) > 1 else "")
            title = title.replace(label, CONFIG.get_ansi_color('search_label') + label + '\x1b[0m')
            output.append(title)

            for idx, match in enumerate(matches):
                for line in match:
                    line = re.sub(rf'({largs.query})',
                                  CONFIG.get_ansi_color('search_query') + r'\1' + '\x1b[0m',
                                  line, flags=re_flags)
                    output.append(f"[{idx+1}]\t".expandtabs(8) + line)

        print('\n'.join(output), file=out)
        return (hits, labels)

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Search command triggered from TUI.')
        tui.viewport.clear()
        # handle input via prompt
        command, results = tui.execute_command('search', out=tui.viewport.buffer)
        if tui.viewport.buffer.lines and results is not None:
            hits, labels = results
            tui.STATE.mode = 'search'
            cur_y, _ = tui.viewport.pad.getyx()
            tui.STATE.previous_line = cur_y
            tui.viewport.buffer.split()
            LOGGER.debug('Applying selection highlighting in search results.')
            for label in labels:
                if label not in tui.selection:
                    continue
                # we match the label including its 'search_label' highlight to ensure that we really
                # only match this specific occurrence of whatever the label may be
                tui.viewport.buffer.replace(range(tui.viewport.buffer.height),
                                            re.escape(CONFIG.get_ansi_color('search_label'))
                                            + label + re.escape('\x1b[0m'),
                                            CONFIG.get_ansi_color('search_label') +
                                            CONFIG.get_ansi_color('selection')
                                            + label + '\x1b[0m\x1b[0m')
            LOGGER.debug('Populating viewport with search results.')
            tui.viewport.view(ansi_map=tui.ANSI_MAP)
            # reset current cursor position
            LOGGER.debug('Resetting cursor position to top.')
            tui.STATE.top_line = 0
            tui.STATE.current_line = 0
            # update top statusbar
            tui.STATE.topstatus = "CoBib v{} - {} hit{}".format(__version__, hits,
                                                                "s" if hits > 1 else "")
            tui.statusbar(tui.topbar, tui.STATE.topstatus)
            tui.STATE.inactive_commands = ['Add', 'Filter', 'Sort']
        elif command[1:]:
            msg = f"No search hits for '{shlex.join(command[1:])}'!"
            LOGGER.info(msg)
            tui.prompt_print(msg)
            tui.viewport.update_list()



"""CoBib add command."""

import argparse
import logging
import sys
from collections import OrderedDict

from cobib.config import CONFIG
from cobib.database import read_database, write_database
from cobib.parser import Entry
from .base_command import ArgumentParser, Command
from .edit import EditCommand

LOGGER = logging.getLogger(__name__)


class AddCommand(Command):
    """Add Command."""

    name = 'add'

    def execute(self, args, out=sys.stdout):
        """Add new entry.

        Adds new entries to the database.

        Args: See base class.
        """
        LOGGER.debug('Starting Add command.')
        parser = ArgumentParser(prog="add", description="Add subcommand parser.")
        parser.add_argument("-l", "--label", type=str,
                            help="the label for the new database entry")
        parser.add_argument("-f", "--file", type=str, nargs="+", action="extend",
                            help="files associated with this entry")
        group_add = parser.add_mutually_exclusive_group()
        group_add.add_argument("-a", "--arxiv", type=str,
                               help="arXiv ID of the new references")
        group_add.add_argument("-b", "--bibtex", type=argparse.FileType('r'),
                               help="BibLaTeX bibliographic data")
        group_add.add_argument("-d", "--doi", type=str,
                               help="DOI of the new references")
        group_add.add_argument("-i", "--isbn", type=str,
                               help="ISBN of the new references")
        parser.add_argument("tags", nargs=argparse.REMAINDER,
                            help="A list of space-separated tags to associate with this entry." +
                            "\nYou can use quotes to specify tags with spaces in them.")
        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        new_entries = OrderedDict()

        edit_entries = False
        if largs.bibtex is not None:
            LOGGER.debug("Adding entries from BibLaTeX '%s'.", largs.bibtex)
            new_entries = Entry.from_bibtex(largs.bibtex)
        elif largs.arxiv is not None:
            LOGGER.debug("Adding entries from arXiv '%s'.", largs.arxiv)
            new_entries = Entry.from_arxiv(largs.arxiv)
        elif largs.doi is not None:
            LOGGER.debug("Adding entries from DOI '%s'.", largs.doi)
            new_entries = Entry.from_doi(largs.doi)
        elif largs.isbn is not None:
            LOGGER.debug("Adding entries from ISBN '%s'.", largs.isbn)
            new_entries = Entry.from_isbn(largs.isbn)
        elif largs.label is not None:
            LOGGER.warning("No input to parse. Creating new entry '%s' manually.", largs.label)
            new_entries = {
                largs.label: Entry(largs.label,
                                   {'ID': largs.label,
                                    'ENTRYTYPE': CONFIG.config['FORMAT']['default_entry_type']})
                }
            edit_entries = True
        else:
            LOGGER.error("Neither an input to parse nor a label for manual creation specified!")
            return

        if largs.label is not None:
            assert len(new_entries.values()) == 1
            for value in new_entries.values():
                # logging done by cobib/parser.py
                value.set_label = largs.label
            new_entries = OrderedDict((largs.label, value) for value in new_entries.values())

        if largs.file is not None:
            assert len(new_entries.values()) == 1
            for value in new_entries.values():
                # logging done by cobib/parser.py
                value.set_file = largs.file

        if largs.tags != []:
            assert len(new_entries.values()) == 1
            for value in new_entries.values():
                # logging done by cobib/parser.py
                value.set_tags = largs.tags

        # Write the new entries to the database. This destructively overwrite the variable with a
        # list of labels of the entries which have actually been added to the database.
        new_entries = write_database(new_entries)

        if edit_entries:
            if largs.label not in new_entries:
                msg = f"You tried to add a new entry '{largs.label}' which already exists!\n" \
                    + f"Please use `cobib edit {largs.label}` instead!"
                LOGGER.warning(msg)
            else:
                read_database(fresh=True)
                EditCommand().execute([largs.label])

        self.git(args=vars(largs))

        for label in new_entries:
            msg = f"'{label}' was added to the database."
            print(msg)
            LOGGER.info(msg)

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Add command triggered from TUI.')
        # handle input via prompt
        tui.execute_command('add')
        # update database list
        LOGGER.debug('Updating list after Add command.')
        read_database(fresh=True)
        tui.viewport.update_list()



"""CoBib delete command."""

import argparse
import logging
import os
import sys

from cobib.config import CONFIG
from cobib.database import read_database
from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class DeleteCommand(Command):
    """Delete Command."""

    name = 'delete'

    def execute(self, args, out=sys.stdout):
        """Delete entries.

        Deletes the entries from the database.

        Args: See base class.
        """
        LOGGER.debug('Starting Delete command.')
        parser = ArgumentParser(prog="delete", description="Delete subcommand parser.")
        parser.add_argument("labels", type=str, nargs='+', help="labels of the entries")

        if not args:
            parser.print_usage(sys.stderr)
            sys.exit(1)

        try:
            largs = parser.parse_args(args)
        except argparse.ArgumentError as exc:
            print("{}: {}".format(exc.argument_name, exc.message), file=sys.stderr)
            return

        conf_database = CONFIG.config['DATABASE']
        file = os.path.expanduser(conf_database['file'])
        with open(file, 'r') as bib:
            lines = bib.readlines()
        entry_to_be_deleted = False
        current_label = None
        deleted_entries = []
        buffer = []
        for line in lines:
            if any([line.startswith(label + ':') for label in largs.labels]):
                current_label = line[:line.find(':')]
                LOGGER.debug('Entry "%s" found. Starting to remove lines.', current_label)
                entry_to_be_deleted = True
                buffer.pop()
                continue
            if entry_to_be_deleted and line.startswith('...'):
                LOGGER.debug('Reached end of entry "%s".', current_label)
                deleted_entries.append(current_label)
                entry_to_be_deleted = False
                continue
            if not entry_to_be_deleted:
                buffer.append(line)
        with open(file, 'w') as bib:
            for line in buffer:
                bib.write(line)

        self.git(args=vars(largs))

        for label in deleted_entries:
            msg = f"'{label}' was removed from the database."
            print(msg)
            LOGGER.info(msg)

    @staticmethod
    def tui(tui):
        """See base class."""
        LOGGER.debug('Delete command triggered from TUI.')
        if tui.selection:
            # use selection for command
            labels = list(tui.selection)
            tui.selection.clear()
        else:
            # get current label
            label, _ = tui.viewport.get_current_label()
            labels = [label]
        # delete selected entry
        tui.execute_command(['delete'] + labels, skip_prompt=True)
        # update database list
        LOGGER.debug('Updating list after Delete command.')
        read_database(fresh=True)
        tui.viewport.update_list()
        # if cursor line is below buffer height, move it one line back up
        if tui.STATE.current_line >= tui.viewport.buffer.height:
            tui.STATE.current_line -= 1



"""CoBib parsing module."""

from collections import OrderedDict
import json
import logging
import os
import re
import subprocess
import sys

from bs4 import BeautifulSoup
from pylatexenc.latexencode import UnicodeToLatexEncoder
from ruamel import yaml
import bibtexparser
import requests

from cobib.config import CONFIG

LOGGER = logging.getLogger(__name__)

# API and HEADER settings according to this resource: https://crosscite.org/docs.html
DOI_URL = "https://doi.org/"
DOI_HEADER = {'Accept': "application/x-bibtex"}
# DOI regex used for matching DOIs
DOI_REGEX = r'(10\.[0-9a-zA-Z]+\/(?:(?!["&\'])\S)+)\b'
# arXiv URL according to docs from here https://arxiv.org/help/oa
ARXIV_URL = "https://export.arxiv.org/api/query?id_list="
# ISBN regex used for matching ISBNs (adapted from https://github.com/xlcnd/isbnlib)
ISBN_REGEX = re.compile(r'97[89]{1}(?:-?\d){10}|\d{9}[0-9X]{1}|'
                        r'[-0-9X]{10,16}', re.I | re.M | re.S)
# ISBN-API: https://openlibrary.org/dev/docs/api/books
ISBN_URL = "https://openlibrary.org/api/books?bibkeys=ISBN:"
# biblatex default types and required values taken from their docs: https://ctan.org/pkg/biblatex
BIBTEX_TYPES = {
    'article': ['author', 'title', 'journal', 'year'],
    'book': ['author', 'title', 'year'],
    'collection': ['editor', 'title', 'year'],
    'proceedings': ['title', 'year'],
    'report': ['author', 'title', 'type', 'institution', 'year'],
    'thesis': ['author', 'title', 'type', 'institution', 'year'],
    'unpublished': ['author', 'title', 'year']
    }


class Entry:
    """Bibliography entry class.

    Handles everything ranging from field manipulation over format conversion to filter matching.
    """
    class YamlDumper(yaml.YAML):
        """Wrapper class for YAML dumping."""

        # pylint: disable=arguments-differ,inconsistent-return-statements
        def dump(self, data, stream=None, **kw):
            """See base class."""
            inefficient = False
            if stream is None:
                inefficient = True
                stream = yaml.compat.StringIO()
            yaml.YAML.dump(self, data, stream, **kw)
            if inefficient:
                return stream.getvalue()

    def __init__(self, label, data, suppress_warnings=True):
        """Initializes the Entry object.

        Args:
            label (str): Database Id used for this entry.
            data (dict): Dictionary of fields specifying this entry.
            suppress_warnings (bool): if True, suppresses warnings.
        """
        label = str(label)
        LOGGER.debug('Initializing entry: %s', label)
        self._label = label
        self.data = data.copy()
        self.escape_special_chars(suppress_warnings)
        month_type = CONFIG.config['FORMAT'].get('month', None)
        if month_type:
            self.convert_month(month_type)
        if self.data['ID'] != self._label:
            # sanity check for matching label and ID
            LOGGER.warning("Mismatching label '%s' and ID '%s'. Overwriting ID with label.",
                           self._label, self.data['ID'])
            self.set_label = self._label

    def __repr__(self):
        """Returns the entry in its bibtex format."""
        return self.to_bibtex()

    @property
    def label(self):
        """Returns the database Id of this entry."""
        return self._label

    @label.setter
    def set_label(self, label):
        """Sets the database Id of this entry."""
        LOGGER.debug("Changing the label '%s' to '%s'.", self.label, label)
        self._label = label
        LOGGER.debug("Changing the ID '%s' to '%s'.", self.data['ID'], label)
        self.data['ID'] = label

    @property
    def tags(self):
        """Returns the tags of this entry."""
        return self.data.get('tags', None)

    @tags.setter
    def set_tags(self, tags):
        """Sets the tags of this entry."""
        self.data['tags'] = ''.join(tag.strip('+')+', ' for tag in tags).strip(', ')
        LOGGER.debug("Adding the tags '%s' to '%s'.", self.data['tags'], self.label)

    @property
    def file(self):
        """Returns the associated file of this entry."""
        return self.data.get('file', None)

    @file.setter
    def set_file(self, file):
        """Sets the associated file of this entry."""
        if isinstance(file, list):
            file = ', '.join([os.path.abspath(f) for f in file])
        else:
            file = os.path.abspath(file)
        self.data['file'] = file
        LOGGER.debug("Adding '%s' as the file to '%s'.", self.data['file'], self.label)

    def convert_month(self, type_):
        """Converts the month into the specified type.

        The month field of an entry may be stored either in string or number format. This function
        is used to convert between the two options.

        Args:
            type_ (str): may be either 'str' or 'int' indicating the format of the month field
        """
        month = self.data.get('month', None)
        if month is None:
            return
        try:
            month = int(month)
        except ValueError:
            pass
        if type(month).__name__ != type_:
            LOGGER.debug('Converting month type for %s', self.label)
            months = ['jan', 'feb', 'mar', 'apr', 'may', 'jun',
                      'jul', 'aug', 'sep', 'oct', 'nov', 'dec']
            if isinstance(month, str):
                self.data['month'] = str(months.index(month)+1)
            elif isinstance(month, int):
                self.data['month'] = months[month-1]

    def escape_special_chars(self, suppress_warnings=True):
        """Escapes special characters.

        Special characters should be escaped to ensure proper rendering in LaTeX documents. This
        function leverages the existing implementation of the pylatexenc module.

        Args:
            suppress_warnings (bool): if True, suppresses warnings.
        """
        enc = UnicodeToLatexEncoder(non_ascii_only=True,
                                    replacement_latex_protection='braces-all',
                                    unknown_char_policy='keep',
                                    unknown_char_warning=not suppress_warnings or
                                    LOGGER.isEnabledFor(10))  # 10 = DEBUG logging level
        for key, value in self.data.items():
            if key in ('ID', 'file'):
                # do NOT these fields and keep any special characters
                self.data[key] = value
                continue
            if isinstance(value, str):
                self.data[key] = enc.unicode_to_latex(value)

    def matches(self, _filter, _or):
        """Check whether the filter matches.

        CoBib provides an extensive filtering implementation. The filter is specified in the form
        of a dictionary whose keys consist of pairs of (str, bool) entries where the string
        indicates the field to match against and the boolean whether a positive (true) or negative
        (false) match is required. The value obviously refers to what needs to be matched.

        Args:
            _filter (dict): dictionary describing the filter as explained above.
            _or (bool): boolean indicating whether logical OR (true) or AND (false) are used to
                        combine multiple filter items.

        Returns:
            Boolean indicating whether this entry matches the filter.
        """
        LOGGER.debug('Checking whether entry %s matches.', self.label)
        match_list = []
        for key, values in _filter.items():
            if key[0] not in self.data.keys():
                match_list.append(not key[1])
                continue
            for val in values:
                if val not in self.data[key[0]]:
                    match_list.append(not key[1])
                else:
                    match_list.append(key[1])
        if _or:
            return any(m for m in match_list)
        return all(m for m in match_list)

    def search(self, query, context=1, ignore_case=False):
        """Search entry contents for query string.

        The search will try its best to recursively query all the data associated with this entry
        for the given query string.

        Args:
            query (str): text to search for.
            context (int): number of context lines to provide for each match.
            ignore_case (bool): if True, ignore case when searching.

        Returns:
            A list of lists containing the context for each match associated with this entry.
        """
        LOGGER.debug('Searching entry %s for %s.', self.label, query)
        matches = []
        bibtex = str(self).split('\n')
        re_flags = re.IGNORECASE if ignore_case else 0
        for idx, line in enumerate(bibtex):
            if re.search(rf'{query}', line, flags=re_flags):
                # add new match
                matches.append([])
                # upper context
                for string in bibtex[max(idx-context, 0):min(idx+1, len(bibtex))]:
                    if not re.search(rf'{query}', string, flags=re_flags):
                        matches[-1].append(string)
                # matching line itself
                matches[-1].append(line)
                # lower context
                for string in bibtex[max(idx+1, 0):min(idx+context+1, len(bibtex))]:
                    if not re.search(rf'{query}', string, flags=re_flags):
                        matches[-1].append(string)
                    else:
                        break

        if self.file and os.path.exists(self.file):
            grep_prog = CONFIG.config['DATABASE'].get('grep')
            LOGGER.debug('Searching associated file %s with %s', self.file, grep_prog)
            grep = subprocess.Popen([grep_prog, f'-C{context}', query, self.file],
                                    stdout=subprocess.PIPE)
            # extract results
            results = grep.stdout.read().decode().split('--')
            for match in results:
                if match:
                    matches.append([line.strip() for line in match.split('\n') if line.strip()])

        return matches

    def to_bibtex(self):
        """Returns the entry in biblatex format."""
        database = bibtexparser.bibdatabase.BibDatabase()
        database.entries = [self.data]
        LOGGER.debug('Converting entry %s to BibTex format.', self.label)
        return bibtexparser.dumps(database)

    def to_yaml(self):
        """Returns the entry in YAML format."""
        yml = Entry.YamlDumper()
        yml.explicit_start = True
        yml.explicit_end = True
        LOGGER.debug('Converting entry %s to YAML format.', self.label)
        return yml.dump({self._label: dict(sorted(self.data.items()))})

    @staticmethod
    def from_bibtex(file, string=False):
        """Creates a new bibliography from a BibLaTex source file.

        Args:
            file (str or file): string with BibLaTex data or path to the BibLaTex file.
            string (bool, optional): indicates whether the file argument is of string or file type.

        Returns:
            An OrderedDict containing the bibliography as per the provided BibLaTex data.
        """
        bparser = bibtexparser.bparser.BibTexParser()
        bparser.ignore_nonstandard_types = CONFIG.config['DATABASE'].getboolean(
            'ignore_non_standard_types', False)
        if string:
            LOGGER.debug('Loading BibTex string: %s.', file)
            database = bibtexparser.loads(file, parser=bparser)
        else:
            LOGGER.debug('Loading BibTex data from file: %s.', file)
            database = bibtexparser.load(file, parser=bparser)
        bib = OrderedDict()
        for entry in database.entries:
            bib[entry['ID']] = Entry(entry['ID'], entry, suppress_warnings=False)
        return bib

    @staticmethod
    def from_yaml(file):
        """Creates a new bibliography from a YAML source file.

        Args:
            file (file): path to YAML file from which to load database.

        Returns:
            An OrderedDict containing the bibliography as per the provided YAML file.
        """
        bib = OrderedDict()
        LOGGER.debug('Loading YAML data from file: %s.', file)
        with open(file, 'r') as database:
            for entry in yaml.safe_load_all(database):
                for label, data in entry.items():
                    bib[label] = Entry(label, data)
        return bib

    @staticmethod
    def from_doi(doi):
        """Queries the bibtex source for a given DOI.

        Args:
            doi (str): DOI for which to obtain the bibtex data.

        Returns:
            An OrderedDict containing the bibliographic data of the provided DOI.
        """
        try:
            assert re.match(DOI_REGEX, doi)
        except AssertionError:
            msg = f"'{doi}' is not a valid DOI."
            LOGGER.warning(msg)
            print(msg, file=sys.stderr)
            return {}
        LOGGER.info('Gathering BibTex data for DOI: %s.', doi)
        try:
            page = requests.get(DOI_URL+doi, headers=DOI_HEADER, timeout=10)
        except requests.exceptions.ReadTimeout as exc:
            msg = 'The DOI API timed out: ' + str(exc)
            LOGGER.warning(msg)
            print(msg, file=sys.stderr)
            return {}
        return Entry.from_bibtex(page.text, string=True)

    @staticmethod
    def from_arxiv(arxiv):
        """Queries the bibtex source for a given arxiv ID.

        Args:
            arxiv (str): arXiv ID for which to obtain the bibtex data.

        Returns:
            An OrderedDict containing the bibliographic data of the provided arXiv ID.
        """
        LOGGER.info('Gathering BibTex data for arXiv ID: %s.', arxiv)
        page = requests.get(ARXIV_URL+arxiv)
        xml = BeautifulSoup(page.text, features='html.parser')
        if xml.feed.entry.title.contents[0] == 'Error':
            msg = 'The arXiv API returned the following error: ' + \
                  xml.feed.entry.summary.contents[0]
            LOGGER.warning(msg)
            print(msg, file=sys.stderr)
            return {}
        entry = {}
        entry['archivePrefix'] = 'arXiv'
        for key in xml.feed.entry.findChildren(recursive=False):
            if key.name == 'arxiv:doi':
                entry['doi'] = str(key.contents[0])
            elif key.name == 'id':
                entry['arxivid'] = str(key.contents[0]).replace('http://arxiv.org/abs/', '')
                entry['eprint'] = str(key.contents[0])
            elif key.name == 'primary_category':
                entry['primaryClass'] = str(key.attrs['term'])
            elif key.name == 'published':
                # The year must also be stored as a string for compatibility reasons with
                # bibtexparser. However, we perform a conversion to an integer first, to ensure that
                # the year can actually be represented as such.
                entry['year'] = str(int(key.contents[0].split('-')[0]))
                if 'ID' in entry.keys():
                    entry['ID'] = entry['ID'] + str(entry['year'])
                else:
                    entry['ID'] = str(entry['year'])
            elif key.name == 'title':
                entry['title'] = re.sub(r'\s+', ' ', key.contents[0].strip().replace('\n', ' '))
            elif key.name == 'author':
                if 'author' not in entry.keys():
                    first = True
                    entry['author'] = ''
                name = [n.contents[0] for n in key.findChildren()][0]
                if first:
                    if 'ID' in entry.keys():
                        entry['ID'] = name.split()[-1] + entry['ID']
                    else:
                        entry['ID'] = name.split()[-1]
                    first = False
                entry['author'] += '{} and '.format(name)
            elif key.name == 'summary':
                entry['abstract'] = re.sub(r'\s+', ' ', key.contents[0].strip().replace('\n', ' '))
            else:
                LOGGER.warning("The key '%s' of this arXiv entry is not being processed!", key.name)
        if 'doi' in entry.keys():
            entry['ENTRYTYPE'] = 'article'
        else:
            entry['ENTRYTYPE'] = 'unpublished'
        # strip last 'and' from author field
        entry['author'] = entry['author'][:-5]
        bib = OrderedDict()
        bib[entry['ID']] = Entry(entry['ID'], entry)
        return bib

    @staticmethod
    def from_isbn(isbn):
        """Queries the bibtex source for a given ISBN.

        Args:
            isbn (str): ISBN for which to obtain the bibtex data.

        Returns:
            An OrderedDict containing the bibliographic data of the provided ISBN.
        """
        assert re.match(ISBN_REGEX, isbn)
        LOGGER.info('Gathering BibTex data for ISBN: %s.', isbn)
        isbn_plain = ''.join([i for i in isbn if i.isdigit()])
        page = requests.get(ISBN_URL+isbn_plain+'&jscmd=data&format=json', timeout=5)
        contents = dict(json.loads(page.content))
        if not contents:
            msg = f'No data was found for ISBN: {isbn}. If you think this is an error and the ' + \
                  'openlibrary API should provide an entry, please file a bug report. Otherwise' + \
                  ' please try adding this entry manually until more APIs are available in CoBib.'
            LOGGER.warning(msg)
            print(msg, file=sys.stderr)
            return {}
        entry = {}
        for key, value in contents[list(contents.keys())[0]].items():
            if key in ['title', 'url']:
                entry[key] = value
            elif key == 'number_of_pages':
                # we explicitly convert to a string to prevent type errors in the bibtexparser
                entry['pages'] = str(value)
            elif key == 'publish_date':
                entry['date'] = value
                try:
                    entry['year'] = re.search(r'\d{4}', value).group()
                    if 'ID' in entry.keys():
                        entry['ID'] += str(entry['year'])
                    else:
                        entry['ID'] = str(entry['year'])
                except AttributeError:
                    pass
            elif key == 'authors':
                if 'ID' in entry.keys():
                    entry['ID'] = value[0]['name'].split()[-1] + entry['ID']
                else:
                    entry['ID'] = value[0]['name'].split()[-1]
                entry['author'] = ' and'.join([a['name'] for a in value])
            elif key == 'publishers':
                entry['publisher'] = ' and'.join([a['name'] for a in value])
        # set entry-type do 'book'
        entry['ENTRYTYPE'] = 'book'
        bib = OrderedDict()
        bib[entry['ID']] = Entry(entry['ID'], entry)
        return bib
